Package: tceper
Title: Access the 'Open Data API' of Pernambuco Court of Accounts
Version: 0.1.4
Authors@R: c(
    person("Andre", "Leite", email = "leite@castlab.org", role = c("aut", "cre")),
    person("Marcos", "Wasilew", email = "marcos.wasilew@gmail.com", role = "aut"),
    person("Hugo", "Vasconcelos", email = "hugo.vasconcelos@ufpe.br", role = "aut"),
    person("Carlos", "Amorin", email = "carlos.agaf@ufpe.br", role = "aut"),
    person("Diogo", "Bezerra", email = "diogo.bezerra@ufpe.br", role = "aut"))
Description: An R interface to the 'Open Data API' of the Tribunal de Contas do
    Estado de Pernambuco (TCE-PE), the Court of Accounts of the State of
    Pernambuco, Brazil. Provides tidy, ready-to-use functions to query public
    data on revenues, expenditures, commitments, procurement, contracts,
    agreements, public works, legal processes, personnel and reference tables
    for all state and municipal government entities in Pernambuco. All results
    are returned as tibbles with column names converted to 'snake_case' by
    default. Uses 'httr2' for HTTP requests and 'cli' for user-friendly
    messages. See <https://sistemas.tcepe.tc.br/DadosAbertos/> for the API
    documentation.
License: MIT + file LICENSE
URL: https://strategicprojects.github.io/tceper/,
        https://github.com/StrategicProjects/tceper
BugReports: https://github.com/StrategicProjects/tceper/issues
Encoding: UTF-8
Language: en
Depends: R (>= 4.1.0)
Imports: cli (>= 3.6.0), httr2 (>= 1.0.0), jsonlite, rlang (>= 1.1.0),
        tibble, purrr, janitor
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, dplyr, stringr,
        ggplot2
Config/testthat/edition: 3
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-28 20:33:08 UTC; leite
Author: Andre Leite [aut, cre],
  Marcos Wasilew [aut],
  Hugo Vasconcelos [aut],
  Carlos Amorin [aut],
  Diogo Bezerra [aut]
Maintainer: Andre Leite <leite@castlab.org>
Repository: CRAN
Date/Publication: 2026-06-01 14:20:02 UTC
Built: R 4.6.0; ; 2026-06-01 15:27:23 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,tce_fields_tbl)
S3method(print,tce_params_tbl)
export(tce_agreements)
export(tce_benefit_types)
export(tce_bid_details)
export(tce_bid_documents)
export(tce_bid_legal_basis)
export(tce_bid_modalities)
export(tce_bid_stages)
export(tce_bid_statuses)
export(tce_bids)
export(tce_budget_statuses)
export(tce_budgeted_revenues)
export(tce_cache_clear)
export(tce_cache_info)
export(tce_catalog)
export(tce_commitment_liquidations)
export(tce_commitment_payments)
export(tce_commitment_summary)
export(tce_consideranda)
export(tce_contract_amendments)
export(tce_contract_documents)
export(tce_contract_items)
export(tce_contracts)
export(tce_creditor_types)
export(tce_debts_fines)
export(tce_determinations)
export(tce_endpoint)
export(tce_entities)
export(tce_fields)
export(tce_funding_sources)
export(tce_inactivation_reasons)
export(tce_legal_instrument_documents)
export(tce_legal_instrument_items)
export(tce_legal_instrument_participants)
export(tce_legal_instruments)
export(tce_municipal_creditor_types)
export(tce_municipal_entities)
export(tce_municipal_expenditures)
export(tce_municipal_revenues)
export(tce_municipal_transfers)
export(tce_municipalities)
export(tce_object_characteristics)
export(tce_object_classifications)
export(tce_object_nature)
export(tce_outcomes)
export(tce_params)
export(tce_payroll_types)
export(tce_person_creditor_types)
export(tce_processes)
export(tce_procurement_process_budget)
export(tce_procurement_process_documents)
export(tce_procurement_process_participants)
export(tce_procurement_processes)
export(tce_public_works)
export(tce_public_works_audits)
export(tce_public_works_contractors)
export(tce_recommendations)
export(tce_reference_codes)
export(tce_reference_dates)
export(tce_reference_sources)
export(tce_remessa_works)
export(tce_remessa_works_execution)
export(tce_remessa_works_geometry)
export(tce_request)
export(tce_retirement_outcomes)
export(tce_sanctions)
export(tce_school_transport)
export(tce_servants)
export(tce_special_accountability)
export(tce_spending_limits)
export(tce_state_commitment_items)
export(tce_state_creditor_types)
export(tce_state_entities)
export(tce_state_expenditures)
export(tce_state_price_comparison)
export(tce_state_revenues)
export(tce_sub_units)
export(tce_suppliers)
export(tce_update_indices)
importFrom(rlang,"%||%")
importFrom(rlang,.data)
importFrom(rlang,.env)

# tceper <a href="https://github.com/StrategicProjects/tceper"><img src="man/figures/logo.svg" align="right" height="139" alt="tceper website" /></a>

<!-- badges: start -->
[![R-CMD-check](https://img.shields.io/badge/R--CMD--check-passing-brightgreen)](https://github.com/StrategicProjects/tceper) 
![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/tceper) 
![CRAN Downloads](https://cranlogs.r-pkg.org/badges/grand-total/tceper) 
![License](https://img.shields.io/badge/license-MIT-darkviolet.svg) 
![](https://img.shields.io/badge/devel%20version-0.1.3-blue.svg)
<!-- badges: end -->

`tceper` is an R client for the [TCE-PE](https://www.tce.pe.gov.br/) (Tribunal de Contas do Estado de Pernambuco) Open Data API.

The package wraps **71 API endpoints** into user-friendly functions that accept `snake_case` parameter names and return tibbles. A built-in catalog lets you discover endpoints, inspect their input parameters and output fields — all without leaving R.


<table class="warning-banner"><tr><td>
&#x26A0;&#xFE0F; <strong class="warning-title">Warning</strong><br>
The TCE-PE API (<code>sistemas.tcepe.tc.br</code>) is <strong>geo-restricted</strong> and only
accessible from Brazilian IP addresses. If you are outside Brazil, API
queries will time out. The discovery functions <code>tce_catalog()</code>,
<code>tce_params()</code> and <code>tce_fields()</code> work offline anywhere, as they read
from the built-in catalog.
</td></tr></table>

<table class="important-banner"><tr><td>
&#x2755;️ <strong class="important-title">Disclaimer</strong><br>
This package acts as a wrapper for a Brazilian public API provided by the 
Tribunal de Contas do Estado de Pernambuco (TCE-PE), which is the institution 
responsible for the data. To maintain consistency with R package development 
standards, all wrapper functions are named in English. However, because the 
source API is natively in Portuguese, you will interact with Portuguese terms 
when passing arguments or parsing the responses. For example: you may use the 
function `tce_state_revenues()`⁠, but you need to pass the parameters in
Portuguese, such as `AnoReferencia = 2025`⁠. You can find the original list of
endpoints and their respective parameters in the [official API documentation](https://sistemas.tcepe.tc.br/DadosAbertos/Exemplo!listar).
</td></tr></table>

## Installation

``` r
# install.packages("remotes")
remotes::install_github("StrategicProjects/tceper")
```

## Quick start

```r
library(tceper)

# 1. Discover endpoints
tce_catalog()
tce_catalog(search = "contrat")

# 2. Inspect parameters and output fields
tce_params("Contratos")
tce_fields("Contratos")

# 3. Query (use snake_case or the original API names)
tce_contracts(codigo_efisco_ug = "510101")
```

## Explore the API

Before querying any endpoint, you can explore the built-in catalog directly from R.

### 1. Browse the catalog

```r
tce_catalog()
#> # A tibble: 71 × 4
#>    endpoint               group          title                         url
#>    <chr>                  <chr>          <chr>                         <chr>
#>  1 ReceitasEstaduais      Receitas       Receitas Estaduais            …
#>  2 ReceitasMunicipais     Receitas       Receitas Municipais           …
#>  …

tce_catalog(search = "licit")
```

### 2. Inspect input parameters

```r
tce_params("Contratos")
#> ── Contratos has 23 parameters: ──────────────────────────────
#> ℹ unidade_gestora, unidade_orcamentaria, esfera, …
#> ── Dictionary ────────────────────────────────────────────────
#> # A tibble: 23 × 5
#>    api_name      r_name          required type      description
#>    <chr>         <chr>           <lgl>    <chr>     <chr>
#>  1 UnidadeGesto… unidade_gestor… FALSE    character Unidade Gestora
#>  …
```

### 3. Inspect output fields

```r
tce_fields("Contratos")
```

### 4. Query

Use `snake_case` names (from `r_name`) or the original API names — both work:

```r
# These are equivalent:
tce_contracts(codigo_efisco_ug = "510101")
tce_contracts(CodigoEfiscoUG = "510101")

# Add more filters
tce_contracts(codigo_efisco_ug = "510101", ano_contrato = "2025")
```

## Verbose mode

When `verbose = TRUE`, the package prints the final API URL and helper commands for inspecting the endpoint:

```r
tce_contracts(codigo_efisco_ug = "510101", verbose = TRUE)
#> ℹ API URL: https://sistemas.tcepe.tc.br/DadosAbertos/Contratos!json?CodigoEfiscoUG=510101
#> ℹ To inspect this endpoint:
#>     - Input parameters: tce_params("Contratos")
#>     - Output fields: tce_fields("Contratos")
```

Enable globally:

```r
options(tceper.verbose = TRUE)
```

## Cache

All 71 wrapper functions cache results in memory, keyed by endpoint + parameters. Default TTL is 1 hour.

```r
tce_contracts(codigo_efisco_ug = "510101")       # hits the API
tce_contracts(codigo_efisco_ug = "510101")       # cache hit (instant)
tce_contracts(codigo_efisco_ug = "510102")       # different key → hits the API

tce_contracts(codigo_efisco_ug = "510101", cache = FALSE)  # force fresh

tce_cache_info()    # inspect cached entries
tce_cache_clear()   # clear all
```

## Parameter validation

The package validates query parameters against the catalog. If you pass a parameter that doesn't exist for that endpoint, it aborts with a helpful error listing the allowed parameters:

```r
tce_contracts(xyz = "foo")
#> ✖ Unknown query parameter(s) for endpoint Contratos: xyz
#> ℹ Allowed parameters:
#>   • unidade_gestora (UnidadeGestora)
#>   • codigo_efisco_ug (CodigoEfiscoUG)
#>   …
```

## Options

| Option | Default | Description |
|---|---|---|
| `tceper.verbose` | `FALSE` | Print final API URL on every call |
| `tceper.progress` | `TRUE` | Show progress messages |
| `tceper.cache_ttl` | `3600` | Cache time-to-live in seconds |

```r
options(
  tceper.verbose  = TRUE,
  tceper.progress = FALSE,
  tceper.cache_ttl = 7200
)
```

## API limits

The API returns at most **100,000 records** per request. When this limit
is reached, the package issues a warning. Use filters to narrow your query:

```r
tce_municipal_expenditures(
  codigo_municipio = "P113",
  ano_referencia   = "2025"
)
```

## Direct access

If you prefer to pass API parameter names directly, use `tce_request()`:

```r
tce_request("Contratos", CodigoEfiscoUG = "510101", AnoContrato = "2025")
```

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE,
  purl = FALSE  # The TCE-PE API is geo-restricted (Brazil-only). Skip code
                # extraction so R CMD check does not try to run it on CRAN
                # machines outside Brazil.
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----mapping, eval = TRUE-----------------------------------------------------
library(tceper)

# Catalog with endpoint metadata
catalog <- tce_catalog()

# Discover exported wrappers and extract the endpoint used internally.
# We look for calls of the form: tce_cached("Endpoint") or tce_request("Endpoint")
exports <- getNamespaceExports("tceper")

# Keep likely "wrapper" functions
skip <- c(
  "tce_request", "tce_catalog", "tce_endpoint", "tce_params", "tce_fields",
  "tce_cache_info", "tce_cache_clear"
)
wrappers <- setdiff(exports, skip)

extract_endpoint <- function(fn) {
  f <- get(fn, envir = asNamespace("tceper"))
  txt <- paste(deparse(body(f)), collapse = "\n")
  m <- regexpr('tce_(?:cached|request)\\("([^"]+)"', txt, perl = TRUE)
  if (m[1] == -1) return(NA_character_)
  hit <- regmatches(txt, m)
  sub('^tce_(?:cached|request)\\("([^"]+)".*$', "\\1", hit)
}

endpoint <- vapply(wrappers, extract_endpoint, character(1))

mappings <- tibble::tibble(
  r_function = paste0("`", wrappers, "()`"),
  endpoint   = endpoint
) |>
  dplyr::filter(!is.na(endpoint)) |>
  dplyr::left_join(
    catalog |> dplyr::select(endpoint, group, title),
    by = "endpoint"
  ) |>
  dplyr::arrange(group, endpoint, r_function)

mappings

# ---- Endpoint introspection --------------------------------------------------

# NOTE: .tce_catalog() is defined in catalog.R (with in-memory caching).
# Do NOT redefine it here.

# Internal: return list of method objects, tolerant to different JSON shapes
.tce_methods <- function(cat) {
  if (is.null(cat$metodos)) return(list())
  met <- cat$metodos
  # Some serializers may turn it into a data.frame; normalize to list-of-lists
  if (is.data.frame(met)) {
    met <- lapply(seq_len(nrow(met)), function(i) as.list(met[i, , drop = FALSE]))
  }
  met
}

# Internal: normalize endpoint key for matching
.tce_norm_key <- function(x) {
  x <- tolower(as.character(x %||% ""))
  x <- gsub("[^a-z0-9]+", "", x)
  x
}


# Internal: derive an R-friendly parameter name from an API name
.tce_make_r_name <- function(api_name) {
  if (is.null(api_name) || is.na(api_name) || !nzchar(as.character(api_name))) {
    return(NA_character_)
  }
  # janitor is imported; this keeps behavior consistent with clean_names elsewhere
  janitor::make_clean_names(as.character(api_name))
}

#' Get the catalog entry for a given endpoint
#'
#' Returns the full metadata for a single endpoint, including its input
#' parameters (`entrada`) and output fields (`saida`). Useful for
#' programmatic access to endpoint metadata.
#'
#' @param endpoint Endpoint name (e.g., `"Municipios"`, `"LicitacaoUG"`,
#'   `"Contratos"`). Use [tce_catalog()] to see all available endpoints.
#' @return A list with endpoint metadata, including `entrada` (input
#'   parameters) and `saida` (output fields) when available.
#'
#' @seealso [tce_catalog()], [tce_params()], [tce_fields()]
#' @export
#'
#' @examples
#' tce_endpoint("Contratos")
tce_endpoint <- function(endpoint) {
  cat <- .tce_catalog()
  met <- .tce_methods(cat)

  key <- .tce_norm_key(endpoint)
  idx <- which(vapply(met, function(m) .tce_norm_key(m$entidade), character(1)) == key)

  if (length(idx) == 0L) {
    # try partial matches for a friendlier error
    candidates <- vapply(met, function(m) m$entidade %||% "", character(1))
    stop(sprintf("Unknown endpoint '%s'. Use tce_catalog() to list available endpoints.", endpoint), call. = FALSE)
  }
  met[[idx[[1L]]]]
}

#' List available API endpoints
#'
#' Returns a tibble listing all 71 endpoints available in the TCE-PE Open
#' Data API, with their group, description and URL. Use `search` to filter
#' by keyword.
#'
#' @param search Optional character string to filter endpoints by name or
#'   description (case-insensitive). For example, `search = "licit"` returns
#'   all procurement-related endpoints.
#' @return A tibble with columns `endpoint`, `group`, `title`, and `url`.
#'
#' @seealso [tce_params()], [tce_fields()], [tce_endpoint()]
#' @export
#'
#' @examples
#' tce_catalog()
#' tce_catalog(search = "contrat")
#' tce_catalog(search = "municipio")
tce_catalog <- function(search = NULL) {
  cat <- .tce_catalog()
  met <- .tce_methods(cat)
  df <- tibble::tibble(
    endpoint = vapply(met, function(m) m$entidade %||% NA_character_, character(1)),
    group    = vapply(met, function(m) m$grupo %||% NA_character_, character(1)),
    title    = vapply(met, function(m) m$descricao %||% NA_character_, character(1)),
    url      = vapply(met, function(m) m$url %||% NA_character_, character(1))
  )
  if (!is.null(search) && nzchar(search)) {
    s <- tolower(search)
    keep <- grepl(s, tolower(df$endpoint), fixed = TRUE) | grepl(s, tolower(df$title), fixed = TRUE)
    df <- df[keep, , drop = FALSE]
  }
  df
}

#' List input parameters for an endpoint
#'
#' Returns a tibble describing the query parameters accepted by a given
#' endpoint. The `api_name` column shows the exact name expected by the
#' API; the `r_name` column shows the `snake_case` equivalent that you
#' can pass to wrapper functions.
#'
#' @param endpoint Endpoint name (e.g., `"Municipios"`, `"Contratos"`).
#'   Use [tce_catalog()] to see all available endpoints.
#' @return A tibble with columns `api_name`, `r_name`, `required`, `type`,
#'   and `description`. Has a custom print method that summarises the
#'   parameter names at the top.
#'
#' @seealso [tce_fields()], [tce_catalog()], [tce_endpoint()]
#' @export
#'
#' @examples
#' tce_params("Contratos")
#' tce_params("Municipios")
#' tce_params("LicitacaoUG")
tce_params <- function(endpoint) {
  e <- tce_endpoint(endpoint)
  ent <- e$entrada %||% list()

  # Normalize entrada to a list-of-lists
  if (is.data.frame(ent)) {
    ent <- lapply(seq_len(nrow(ent)), function(i) as.list(ent[i, , drop = FALSE]))
  } else if (is.list(ent) && !is.null(names(ent)) && !is.list(ent[[1L]])) {
    # single object -> wrap
    ent <- list(ent)
  }


df <- tibble::tibble(
  api_name     = vapply(ent, function(x) x$api_name %||% x$nome %||% x$parametro %||% NA_character_, character(1)),
  r_name       = NA_character_,
  required     = vapply(ent, function(x) isTRUE(x$required %||% x$obrigatorio %||% FALSE), logical(1)),
  type         = vapply(ent, function(x) x$type %||% x$tipo %||% NA_character_, character(1)),
  description  = vapply(ent, function(x) x$description %||% x$descricao %||% NA_character_, character(1))
)
# Derive r_name when not provided by catalog
df$r_name <- vapply(df$api_name, .tce_make_r_name, character(1))
# Simple default type
df$type[is.na(df$type) | !nzchar(df$type)] <- "character"

df <- structure(df, class = c("tce_params_tbl", class(df)))
attr(df, "endpoint") <- e$entidade %||% endpoint
df
}

#' List output fields for an endpoint
#'
#' Returns a tibble describing the columns returned by a given endpoint.
#' When `clean_names = TRUE` (the default), column names are converted
#' from the API's original names to `snake_case`.
#'
#' @param endpoint Endpoint name (e.g., `"Municipios"`, `"Contratos"`).
#'   Use [tce_catalog()] to see all available endpoints.
#' @return A tibble with columns `name`, `type`, and `description`. Has a
#'   custom print method that summarises the field names at the top.
#'
#' @seealso [tce_params()], [tce_catalog()], [tce_endpoint()]
#' @export
#'
#' @examples
#' tce_fields("Contratos")
#' tce_fields("Municipios")
tce_fields <- function(endpoint) {
  e <- tce_endpoint(endpoint)
  out <- e$saida %||% list()

  if (is.data.frame(out)) {
    out <- lapply(seq_len(nrow(out)), function(i) as.list(out[i, , drop = FALSE]))
  } else if (is.list(out) && !is.null(names(out)) && !is.list(out[[1L]])) {
    out <- list(out)
  }


df <- tibble::tibble(
  name        = vapply(out, function(x) x$name %||% x$campo %||% x$nome %||% NA_character_, character(1)),
  type        = vapply(out, function(x) x$type %||% x$tipo %||% NA_character_, character(1)),
  description = vapply(out, function(x) x$description %||% x$descricao %||% NA_character_, character(1))
)
df$type[is.na(df$type) | !nzchar(df$type)] <- "character"

df <- structure(df, class = c("tce_fields_tbl", class(df)))
attr(df, "endpoint") <- e$entidade %||% endpoint
df
}

# ---- Contracts ---------------------------------------------------------------

#' Contracts
#'
#' Retrieve state and municipal contracts.
#'
#' @param ... Named query parameters passed directly to the API.
#'   Common filters: `CodigoEfiscoUG`, `AnoReferencia`, `CodigoMunicipio`.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to
#'   `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @seealso [tce_params()] to discover available parameters,
#'   [tce_fields()] to see output columns.
#' @export
#' @family procurement
#'
#' @examples
#' \dontrun{
#' # Not run: reaches the live TCE-PE API, which only accepts connections
#' # from Brazilian IP addresses (unreachable from CRAN / most CI).
#' tce_contracts(codigo_efisco_ug = "510101", verbose = TRUE)
#' }
tce_contracts <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(
    tce_cache_key("Contratos", ..., call = .user_call),
    tce_request("Contratos", ..., clean_names = clean_names, verbose = verbose),
    use_cache = cache)
}

#' Contract documents
#'
#' Retrieve documents associated with contracts.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_contract_documents <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("ContratoDocumentos", ..., call = .user_call), tce_request("ContratoDocumentos", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Contract line items
#'
#' Retrieve the individual items that compose a contract.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_contract_items <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("ContratoItemObjeto", ..., call = .user_call), tce_request("ContratoItemObjeto", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Contract amendments
#'
#' Retrieve addenda / amendments (*termos aditivos*) to contracts.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_contract_amendments <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TermoAditivo", ..., call = .user_call), tce_request("TermoAditivo", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Agreements (Convenios)
#'
#' Retrieve state and municipal agreements (*convênios*).
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_agreements <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Convenios", ..., call = .user_call), tce_request("Convenios", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- Bids (Licitações) ------------------------------------------------------

#' Bids
#'
#' Retrieve public procurement bids (*licitações*).
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_bids <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("LicitacaoUG", ..., call = .user_call), tce_request("LicitacaoUG", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Bid details with bidders
#'
#' Retrieve bids with their respective bidders (*licitantes*).
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_bid_details <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("LicitacoesDetalhes", ..., call = .user_call), tce_request("LicitacoesDetalhes", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Bid documents
#'
#' Retrieve documents associated with bids.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_bid_documents <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("LicitacoesDocumentos", ..., call = .user_call), tce_request("LicitacoesDocumentos", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- Procurement lookup tables -----------------------------------------------

#' Bid stages
#'
#' Lookup table of bid process stages.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_bid_stages <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("EstagioLicitacao", ..., call = .user_call), tce_request("EstagioLicitacao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Bid modalities
#'
#' Lookup table of bid modality types.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_bid_modalities <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("ModalidadeLicitacao", ..., call = .user_call), tce_request("ModalidadeLicitacao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Bid statuses
#'
#' Lookup table of bid status types.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_bid_statuses <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("SituacaoLicitacao", ..., call = .user_call), tce_request("SituacaoLicitacao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Object characteristics
#'
#' Lookup table of procurement object characteristic types.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_object_characteristics <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("CaracteristicaObjeto", ..., call = .user_call), tce_request("CaracteristicaObjeto", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Object classifications
#'
#' Lookup table of procurement object classification types.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_object_classifications <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("ClassificacaoObjeto", ..., call = .user_call), tce_request("ClassificacaoObjeto", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Bid legal basis
#'
#' Lookup table of legal basis types for bids.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_bid_legal_basis <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("FundamentacaoLegalLicitacao", ..., call = .user_call), tce_request("FundamentacaoLegalLicitacao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Object nature types
#'
#' Lookup table of procurement object nature types.
#'
#' @inheritParams tce_contracts
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family procurement
tce_object_nature <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("NaturezaObjeto", ..., call = .user_call), tce_request("NaturezaObjeto", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- RemessaTCEPE lookup tables ----------------------------------------------

#' Benefit types (Remessa)
#'
#' Lookup table of benefit types (RemessaTCEPE system).
#'
#' @param ... Named query parameters passed directly to the API.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_benefit_types <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TipoBeneficio", ..., call = .user_call), tce_request("TipoBeneficio", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Reference sources (Remessa)
#'
#' Lookup table of reference sources for cost-estimate imports.
#'
#' @inheritParams tce_benefit_types
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_reference_sources <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("FonteReferencia", ..., call = .user_call), tce_request("FonteReferencia", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Reference dates (Remessa)
#'
#' Available dates for reference sources in cost-estimate imports.
#'
#' @inheritParams tce_benefit_types
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_reference_dates <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("DataReferencia", ..., call = .user_call), tce_request("DataReferencia", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Reference codes (Remessa)
#'
#' Reference codes for cost-estimate item imports.
#'
#' @inheritParams tce_benefit_types
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_reference_codes <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("CodigoReferencia", ..., call = .user_call), tce_request("CodigoReferencia", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Price update indices (Remessa)
#'
#' Indices for updating prices in cost-estimate spreadsheets.
#'
#' @inheritParams tce_benefit_types
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_update_indices <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("IndiceAtualizacao", ..., call = .user_call), tce_request("IndiceAtualizacao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Budget item statuses (Remessa)
#'
#' Possible statuses for cost-estimate items/lots.
#'
#' @inheritParams tce_benefit_types
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_budget_statuses <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("OrcamentoSituacao", ..., call = .user_call), tce_request("OrcamentoSituacao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- RemessaTCEPE: Legal instruments -----------------------------------------

#' Legal instruments (Remessa)
#'
#' Retrieve legal instruments from RemessaTCEPE.
#'
#' @param ... Named query parameters. Common: `CodigoEfiscoUG`.
#' @inheritParams tce_benefit_types
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_legal_instruments <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_InstrumentoJuridico", ..., call = .user_call), tce_request("Remessa_InstrumentoJuridico", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Legal instrument documents (Remessa)
#'
#' @inheritParams tce_legal_instruments
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_legal_instrument_documents <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_InstrumentoJuridicoDocumento", ..., call = .user_call), tce_request("Remessa_InstrumentoJuridicoDocumento", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Legal instrument items (Remessa)
#'
#' @inheritParams tce_legal_instruments
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_legal_instrument_items <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_InstrumentoJuridicoItens", ..., call = .user_call), tce_request("Remessa_InstrumentoJuridicoItens", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Legal instrument participants (Remessa)
#'
#' @inheritParams tce_legal_instruments
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_legal_instrument_participants <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_InstrumentoJuridicoParticipantes", ..., call = .user_call), tce_request("Remessa_InstrumentoJuridicoParticipantes", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- RemessaTCEPE: Procurement processes -------------------------------------

#' Procurement processes (Remessa)
#'
#' Retrieve procurement processes from RemessaTCEPE.
#'
#' @inheritParams tce_legal_instruments
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_procurement_processes <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_ProcessoContratacao", ..., call = .user_call), tce_request("Remessa_ProcessoContratacao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Procurement process documents (Remessa)
#'
#' @inheritParams tce_legal_instruments
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_procurement_process_documents <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_ProcessoContratacaoDocumento", ..., call = .user_call), tce_request("Remessa_ProcessoContratacaoDocumento", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Procurement process budget (Remessa)
#'
#' @inheritParams tce_legal_instruments
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_procurement_process_budget <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_ProcessoContratacaoOrcamento", ..., call = .user_call), tce_request("Remessa_ProcessoContratacaoOrcamento", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Procurement process participants (Remessa)
#'
#' @inheritParams tce_legal_instruments
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_procurement_process_participants <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_ProcessoContratacaoParticipantes", ..., call = .user_call), tce_request("Remessa_ProcessoContratacaoParticipantes", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- RemessaTCEPE: Public works ----------------------------------------------

#' Public works details (Remessa)
#'
#' @inheritParams tce_legal_instruments
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_remessa_works <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_Obra", ..., call = .user_call), tce_request("Remessa_Obra", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Public works execution (Remessa)
#'
#' @inheritParams tce_legal_instruments
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_remessa_works_execution <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_ObraExecucao", ..., call = .user_call), tce_request("Remessa_ObraExecucao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Public works geometry (Remessa)
#'
#' @inheritParams tce_legal_instruments
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family remessa
tce_remessa_works_geometry <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Remessa_ObraGeometria", ..., call = .user_call), tce_request("Remessa_ObraGeometria", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- Verbose helpers ---------------------------------------------------------

#' Print help commands for inspecting an endpoint (verbose mode)
#' @noRd
.tce_verbose_help <- function(endpoint) {
  cmd_params <- sprintf('tce_params("%s")', endpoint)
  cmd_fields <- sprintf('tce_fields("%s")', endpoint)

  if (requireNamespace("cli", quietly = TRUE)) {
    cli::cli_alert_info("To inspect this endpoint:")
    cli::cli_alert("    - Input parameters: {.code { cmd_params }}")
    cli::cli_alert("    - Output fields: {.code { cmd_fields}}")
  } else {
    message("To inspect this endpoint:")
    message("    - Input parameters: ", cmd_params)
    message("    - Output fields:    ", cmd_fields)
  }
}

# ---- Suppliers ---------------------------------------------------------------

#' Suppliers
#'
#' Retrieve registered suppliers. Partial matching is supported.
#'
#' @param ... Named query parameters passed directly to the API.
#'   Common filters: `CpfCnpj`, `Nome`.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family suppliers
tce_suppliers <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Fornecedores", ..., call = .user_call), tce_request("Fornecedores", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Person creditor types
#'
#' Lookup table of person creditor types.
#'
#' @inheritParams tce_suppliers
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family suppliers
tce_person_creditor_types <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TipoCredorPessoa", ..., call = .user_call), tce_request("TipoCredorPessoa", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Sanctions
#'
#' Retrieve sanctions applied to suppliers.
#'
#' @inheritParams tce_suppliers
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family suppliers
tce_sanctions <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Sancoes", ..., call = .user_call), tce_request("Sancoes", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- Revenues ----------------------------------------------------------------

#' State revenues
#'
#' Retrieve state-level revenue data from the TCE-PE API.
#'
#' @param ... Named query parameters passed directly to the API.
#'   Common filters: `CodigoEfiscoUG`, `AnoReferencia`.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to
#'   `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @seealso [tce_params()] to discover available parameters,
#'   [tce_fields()] to see output columns.
#' @export
#' @family revenues
#'
#' @examples
#' \dontrun{
#' # Not run: reaches the live TCE-PE API, which only accepts connections
#' # from Brazilian IP addresses (unreachable from CRAN / most CI).
#' tce_state_revenues(id_unidade_gestora = "510101")
#' }
tce_state_revenues <- function(...,
                               verbose = getOption("tceper.verbose", FALSE),
                               cache = TRUE,
                               clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(
    tce_cache_key("ReceitasEstaduais", ..., call = .user_call),
    tce_request("ReceitasEstaduais", ..., clean_names = clean_names, verbose = verbose),
    use_cache = cache
  )
}

#' Municipal revenues
#'
#' Retrieve municipal-level revenue data.
#'
#' @inheritParams tce_state_revenues
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family revenues
#'
#' @examples
#' \dontrun{
#' # Not run: reaches the live TCE-PE API, which only accepts connections
#' # from Brazilian IP addresses (unreachable from CRAN / most CI).
#' tce_municipal_revenues(id_unidade_gestora = "384")
#' }
tce_municipal_revenues <- function(...,
                                   verbose = getOption("tceper.verbose", FALSE),
                                   cache = TRUE,
                                   clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(
    tce_cache_key("ReceitasMunicipais", ..., call = .user_call),
    tce_request("ReceitasMunicipais", ..., clean_names = clean_names, verbose = verbose),
    use_cache = cache
  )
}

#' Budgeted revenues (municipal)
#'
#' Retrieve budgeted (forecast) revenue data. Municipal entities only.
#'
#' @inheritParams tce_state_revenues
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family revenues
tce_budgeted_revenues <- function(...,
                                  verbose = getOption("tceper.verbose", FALSE),
                                  cache = TRUE,
                                  clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(
    tce_cache_key("ReceitasPrevistas", ..., call = .user_call),
    tce_request("ReceitasPrevistas", ..., clean_names = clean_names, verbose = verbose),
    use_cache = cache
  )
}

# ---- Print methods for introspection tibbles ---------------------------------

#' Print method for `tce_params()` output
#'
#' @keywords internal
#' @param x An object returned by [tce_params()].
#' @param ... Passed to the next print method.
#' @return `x`, invisibly.
#' @method print tce_params_tbl
#' @export
print.tce_params_tbl <- function(x, ...) {
  ep <- attr(x, "endpoint") %||% "unknown"
  n  <- nrow(x)

  params <- x$r_name
  params_str <- paste(params, collapse = ", ")

  if (requireNamespace("cli", quietly = TRUE)) {
    cli::cli_rule(left = "{ep} has {n} parameter{?s}:")
    cli::cli_inform(c(
      "i" = "{.field {params_str}}."
    ))
    cli::cli_rule(left = "Dictionary")
  } else {
    cat(sprintf("%s has %d parameter(s): %s.\n", ep, n, params_str))
  }

  NextMethod("print", x, ...)
}

#' Print method for `tce_fields()` output
#'
#' @keywords internal
#' @param x An object returned by [tce_fields()].
#' @param ... Passed to the next print method.
#' @return `x`, invisibly.
#' @method print tce_fields_tbl
#' @export
print.tce_fields_tbl <- function(x, ...) {
  ep <- attr(x, "endpoint") %||% "unknown"
  n  <- nrow(x)

  fields <- x$name
  fields_str <- paste(fields, collapse = ", ")

  if (requireNamespace("cli", quietly = TRUE)) {
    cli::cli_rule(left = "{ep} has {n} field{?s}:")
    cli::cli_inform(c(
      "i" = "{.field {fields_str}}."
    ))
    cli::cli_rule(left = "Dictionary")
  } else {
    cat(sprintf("%s has %d field(s): %s.\n", ep, n, fields_str))
  }

  NextMethod("print", x, ...)
}

# ---- Legal processes ---------------------------------------------------------

#' Legal processes
#'
#' Retrieve the list of formalized physical and electronic processes.
#'
#' @param ... Named query parameters passed directly to the API.
#'   Common filters: `CodigoEfiscoUG`, `AnoReferencia`, `CodigoMunicipio`.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to
#'   `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family processes
tce_processes <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Processos", ..., call = .user_call), tce_request("Processos", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Determinations
#'
#' Retrieve determinations from adjudicated electronic processes.
#'
#' @inheritParams tce_processes
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family processes
tce_determinations <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Determinacoes", ..., call = .user_call), tce_request("Determinacoes", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Consideranda
#'
#' Retrieve *considerandos* (findings/grounds) from adjudicated processes.
#'
#' @inheritParams tce_processes
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family processes
tce_consideranda <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Considerandos", ..., call = .user_call), tce_request("Considerandos", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Recommendations
#'
#' Retrieve recommendations from adjudicated electronic processes.
#'
#' @inheritParams tce_processes
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family processes
tce_recommendations <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Recomendacoes", ..., call = .user_call), tce_request("Recomendacoes", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Process outcomes
#'
#' Retrieve outcomes of adjudicated electronic processes.
#'
#' @inheritParams tce_processes
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family processes
tce_outcomes <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Resultados", ..., call = .user_call), tce_request("Resultados", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Retirement, pension and reform outcomes
#'
#' Retrieve outcomes from processes related to retirement, pension and
#' reform (*Aposentadoria, Pensão e Reforma*).
#'
#' @inheritParams tce_processes
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family processes
tce_retirement_outcomes <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("ResultadosAPR", ..., call = .user_call), tce_request("ResultadosAPR", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Special accountability outcomes
#'
#' Retrieve outcomes from special accountability processes
#' (*Tomada de Conta Especial*).
#'
#' @inheritParams tce_processes
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family processes
tce_special_accountability <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("ResultadosTomadaContaEspecial", ..., call = .user_call), tce_request("ResultadosTomadaContaEspecial", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Debts and fines
#'
#' Retrieve debts and fines imposed.
#'
#' @inheritParams tce_processes
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family processes
tce_debts_fines <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("DebitosMultas", ..., call = .user_call), tce_request("DebitosMultas", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Spending limits
#'
#' Retrieve spending-limit data from government account processes.
#'
#' @inheritParams tce_processes
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family processes
tce_spending_limits <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("DadosLimiteGastos", ..., call = .user_call), tce_request("DadosLimiteGastos", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- Cache infrastructure ----------------------------------------------------

#' Cache environment (package-internal)
#' @noRd
tce_cache_env <- new.env(parent = emptyenv())

#' Retrieve a cached result or execute and cache the expression
#' @noRd
tce_cached <- function(key, expr, use_cache = TRUE) {
  if (!use_cache) return(expr)

  ttl <- getOption("tceper.cache_ttl", 3600)

  if (exists(key, envir = tce_cache_env, inherits = FALSE)) {
    entry <- get(key, envir = tce_cache_env, inherits = FALSE)
    age <- as.numeric(difftime(Sys.time(), entry$time, units = "secs"))
    if (age < ttl) {
      if (getOption("tceper.progress", TRUE)) {
        nrows <- if (tibble::is_tibble(entry$value)) nrow(entry$value) else NA_integer_
        cli::cli_alert_success(
          "Cache hit ({.val {round(age)}}s old{if (!is.na(nrows)) sprintf(', %s row%s', nrows, if (nrows == 1L) '' else 's') else ''})."
        )
      }
      return(entry$value)
    }
  }

  result <- expr
  assign(key, list(value = result, time = Sys.time()), envir = tce_cache_env)
  result
}

#' Clear the tceper in-memory cache
#'
#' Removes cached API responses so that subsequent calls hit the API again.
#' By default clears everything; use `pattern` to selectively clear entries
#' for specific endpoints.
#'
#' @param pattern Optional regex pattern to clear only matching keys.
#'   If `NULL` (default), clears the entire cache. For example,
#'   `"Contratos"` clears all cached contract queries.
#'
#' @return Invisibly returns the number of entries cleared.
#'
#' @seealso [tce_cache_info()]
#' @export
#'
#' @examples
#' tce_cache_clear()              # clear everything
#' tce_cache_clear("Contratos")   # clear contract queries only
tce_cache_clear <- function(pattern = NULL) {
  keys <- ls(envir = tce_cache_env)
  if (!is.null(pattern)) keys <- grep(pattern, keys, value = TRUE)
  rm(list = keys, envir = tce_cache_env)
  if (length(keys) > 0) {
    cli::cli_alert_success("Cleared {.val {length(keys)}} cached entr{?y/ies}.")
  } else {
    cli::cli_alert_info("No cached entries to clear.")
  }
  invisible(length(keys))
}

#' Show cache status
#'
#' Returns a tibble describing all currently cached entries, including
#' the cache key (endpoint + parameters), row count, timestamp and age.
#' Useful for understanding what is cached and when it will expire.
#'
#' @return A [tibble][tibble::tibble-package] with columns `key`, `rows`,
#'   `cached_at` and `age_secs`, or an empty tibble if the cache is empty.
#'   The cache TTL is controlled by `getOption("tceper.cache_ttl", 3600)`.
#'
#' @seealso [tce_cache_clear()]
#' @export
#'
#' @examples
#' tce_cache_info()
tce_cache_info <- function() {
  keys <- ls(envir = tce_cache_env)
  if (length(keys) == 0) {
    cli::cli_alert_info("Cache is empty.")
    return(tibble::tibble(
      key = character(), rows = integer(),
      cached_at = as.POSIXct(character()), age_secs = numeric()
    ))
  }
  now <- Sys.time()
  entries <- lapply(keys, function(k) {
    entry <- get(k, envir = tce_cache_env, inherits = FALSE)
    nrows <- if (tibble::is_tibble(entry$value)) nrow(entry$value) else NA_integer_
    tibble::tibble(
      key = k, rows = nrows, cached_at = entry$time,
      age_secs = round(as.numeric(difftime(now, entry$time, units = "secs")))
    )
  })
  do.call(rbind, entries)
}

# ---- Expenditures ------------------------------------------------------------

#' State expenditures
#'
#' Retrieve state-level expenditure data.
#'
#' @param ... Named query parameters passed directly to the API.
#'   Common filters: `CodigoEfiscoUG`, `AnoReferencia`.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to
#'   `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @seealso [tce_params()] to discover available parameters,
#'   [tce_fields()] to see output columns.
#' @export
#' @family expenditures
#'
#' @examples
#' \dontrun{
#' # Not run: reaches the live TCE-PE API, which only accepts connections
#' # from Brazilian IP addresses (unreachable from CRAN / most CI).
#' tce_state_expenditures(id_unidade_gestora = "510101")
#' }
tce_state_expenditures <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("DespesasEstaduais", ..., call = .user_call), tce_request("DespesasEstaduais", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Municipal expenditures
#'
#' Retrieve municipal-level expenditure data.
#'
#' @inheritParams tce_state_expenditures
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family expenditures
tce_municipal_expenditures <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("DespesasMunicipais", ..., call = .user_call), tce_request("DespesasMunicipais", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Municipal granted transfers
#'
#' Retrieve inter-unit transfers granted within the same municipality.
#'
#' @inheritParams tce_state_expenditures
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family expenditures
tce_municipal_transfers <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TransferenciasConcedidasMunicipais", ..., call = .user_call), tce_request("TransferenciasConcedidasMunicipais", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Municipal creditor types
#'
#' Lookup table of municipal creditor types.
#'
#' @inheritParams tce_state_expenditures
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family expenditures
tce_municipal_creditor_types <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TipoCredorMunicipal", ..., call = .user_call), tce_request("TipoCredorMunicipal", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' State creditor types
#'
#' Lookup table of state creditor types.
#'
#' @inheritParams tce_state_expenditures
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family expenditures
tce_state_creditor_types <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TipoCredorEstadual", ..., call = .user_call), tce_request("TipoCredorEstadual", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- Commitments (Empenhos) --------------------------------------------------

#' Commitment summary
#'
#' Retrieve original values, reinforcements and cancellations for budget
#' commitments (*empenhos*), for both municipal and state entities.
#'
#' @param ... Named query parameters passed directly to the API.
#'   Common filters: `CodigoEfiscoUG`, `NumeroEmpenho`, `AnoReferencia`.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to
#'   `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family commitments
tce_commitment_summary <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("EmpenhoResumo", ..., call = .user_call), tce_request("EmpenhoResumo", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Commitment liquidations
#'
#' Retrieve liquidation records for budget commitments.
#'
#' @inheritParams tce_commitment_summary
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family commitments
tce_commitment_liquidations <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("EmpenhoLiquidacao", ..., call = .user_call), tce_request("EmpenhoLiquidacao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Commitment payments
#'
#' Retrieve payment records for budget commitments (municipal only).
#'
#' @inheritParams tce_commitment_summary
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family commitments
tce_commitment_payments <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("EmpenhoPagamento", ..., call = .user_call), tce_request("EmpenhoPagamento", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' State commitment line items
#'
#' Retrieve individual line items of state-level commitments.
#'
#' @inheritParams tce_commitment_summary
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family commitments
tce_state_commitment_items <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("ItemEmpenhoEstadual", ..., call = .user_call), tce_request("ItemEmpenhoEstadual", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' State price comparison
#'
#' Retrieve state commitment items with values for price comparison.
#'
#' @inheritParams tce_commitment_summary
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family commitments
tce_state_price_comparison <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("ComparativoPrecoEstado", ..., call = .user_call), tce_request("ComparativoPrecoEstado", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' tceper: Access the Open Data API of TCE-PE
#'
#' An R interface to the Open Data API of the Tribunal de Contas do Estado
#' de Pernambuco (TCE-PE), the Court of Accounts of the State of Pernambuco,
#' Brazil. Provides tidy, ready-to-use functions to query public data on
#' revenues, expenditures, commitments, procurement, contracts, agreements,
#' public works, legal processes, personnel and reference tables for all
#' state and municipal government entities in Pernambuco.
#'
#' @section Access to the API:
#'
#' The TCE-PE API host (`sistemas.tcepe.tc.br`) is **geo-restricted**
#' and only accepts requests from Brazilian IP addresses. Users outside
#' Brazil will see connection or TLS errors. The discovery functions
#' ([tce_catalog()], [tce_endpoint()], [tce_params()], [tce_fields()])
#' are unaffected -- they read from a built-in catalog and work offline
#' from any location.
#'
#' @section Recommended workflow:
#'
#' 1. **Discover** endpoints with [tce_catalog()] (filter with `search =`).
#' 2. **Inspect** input parameters with [tce_params()] and output fields
#'    with [tce_fields()].
#' 3. **Query** using a wrapper function (e.g. [tce_contracts()]) or the
#'    low-level [tce_request()].
#'
#' @section Parameter names:
#'
#' All query parameters are passed through `...` using the names shown
#' by [tce_params()]. You can use either the original API name (e.g.
#' `CodigoEfiscoUG`) or its `snake_case` equivalent (e.g.
#' `codigo_efisco_ug`); the package maps them automatically.
#'
#' @section Options:
#'
#' \describe{
#'   \item{`tceper.verbose`}{If `TRUE`, all functions print the final API
#'     URL. Default: `FALSE`.}
#'   \item{`tceper.progress`}{If `FALSE`, suppresses progress messages.
#'     Default: `TRUE`.}
#'   \item{`tceper.cache_ttl`}{Cache time-to-live in seconds.
#'     Default: `3600` (1 hour).}
#' }
#'
#' @seealso
#' - [tce_catalog()], [tce_params()], [tce_fields()] — endpoint discovery
#' - [tce_request()] — low-level query function
#' - [tce_cache_info()], [tce_cache_clear()] — cache management
#'
#' @keywords internal
#' @aliases tceper-package
"_PACKAGE"

#' @importFrom rlang .data .env %||%
NULL

# ---- Core request engine -----------------------------------------------------

#' Perform a request to the TCE-PE Open Data API
#'
#' @description
#' Low-level function that sends a GET request to the TCE-PE Open Data API
#' and returns the response content as a [tibble][tibble::tibble-package].
#' Higher-level wrapper functions such as [tce_contracts()],
#' [tce_state_expenditures()], etc. call this function internally.
#'
#' You can use `tce_request()` directly to access any API endpoint,
#' including those that do not yet have a dedicated wrapper.
#'
#' @param endpoint A character string with the API method name (e.g.,
#'   `"Contratos"`, `"DespesasEstaduais"`). See
#'   <https://sistemas.tcepe.tc.br/DadosAbertos/Exemplo!listar> for the
#'   full list of available endpoints.
#' @param ... Named query parameters. You can use either the API names
#'   (`CamelCase`/`UPPERCASE`) or the snake_case aliases shown by
#'   [tce_params()]. `NULL` values are silently dropped.
#' @param clean_names If `TRUE` (default), column names are converted to
#'   `snake_case` via [janitor::clean_names()]. Set to `FALSE` to keep the
#'   original column names returned by the API.
#' @param max_tries Maximum number of retry attempts for failed requests
#'   (default: `3`). Passed to [httr2::req_retry()].
#' @param progress If `TRUE`, displays progress messages via [cli]. Defaults
#'   to the value of `getOption("tceper.progress")`, or `TRUE` if unset.
#' @param verbose If `TRUE`, prints the final API URL, parameters, and
#'   response metadata (status, content-type, size, elapsed time). Defaults
#'   to `getOption("tceper.verbose", FALSE)`.
#'
#' @return A [tibble][tibble::tibble-package] with the response content.
#'   Returns a zero-row tibble if no records are found.
#'
#' @details
#' ## API base URL
#' `https://sistemas.tcepe.tc.br/DadosAbertos/`
#'
#' ## Network access
#' The API host only accepts connections from **Brazilian IP addresses**.
#' Calls from outside Brazil are dropped at the network level and surface
#' as a connection or TLS error. The offline discovery functions
#' ([tce_catalog()], [tce_params()], [tce_fields()]) do not require
#' network access and work anywhere.
#'
#' ## URL construction
#' The TCE-PE API uses Struts2-style URLs where `!` invokes a method
#' (e.g. `Contratos!json`). The URL is built with `paste0()` to preserve
#' the literal `!` -- using `httr2::req_url_path_append()` would encode it
#' as `%21`, causing the server to ignore all query parameters.
#'
#' ## Response limit
#' The API returns at most **100,000 records** per request. When this limit
#' is reached, a warning is issued suggesting the use of additional filters.
#'
#' ## Encoding
#' The TCE-PE backend runs on **ISO-8859-1** (Latin-1), not UTF-8. The
#' package handles this transparently in both directions:
#'
#' * **Response body** is decoded as Latin-1 and returned as UTF-8 R
#'   strings, so accented Portuguese characters come through cleanly
#'   (`São`, `Município`, `Conceição`, etc.).
#' * **Query parameters** are transcoded UTF-8 -> Latin-1 before
#'   percent-encoding, so filtering by accented values works as
#'   expected (e.g. `tce_municipalities(municipio = "São José da Coroa Grande")`).
#'   You can pass any R string with accents -- the conversion is
#'   automatic.
#'
#' ## Options
#' - `tceper.progress` -- set to `FALSE` to suppress all progress messages.
#' - `tceper.verbose` -- set to `TRUE` to always print request/response details.
#' - `tceper.cache_ttl` -- cache time-to-live in seconds (default: `3600`).
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # Not run: these examples reach the live TCE-PE API, which only accepts
#' # connections from Brazilian IP addresses and is therefore unreachable
#' # from CRAN check machines and most CI runners.
#'
#' # Direct query to any endpoint
#' tce_request("Contratos", CodigoEfiscoUG = "510101")
#'
#' # Show full request/response details
#' tce_request("Contratos", CodigoEfiscoUG = "510101", verbose = TRUE)
#'
#' # Keep original column names
#' tce_request("Municipios", clean_names = FALSE)
#' }
tce_request <- function(endpoint,
                        ...,
                        clean_names = TRUE,
                        max_tries = 3,
                        progress = getOption("tceper.progress", TRUE),
                        verbose = getOption("tceper.verbose", FALSE)) {
  .user_call <- rlang::caller_call()

  params <- tce_build_params(...)
  .tce_assert_allowed_params(endpoint, params, call = .user_call)
  params <- .tce_map_params(endpoint, params)

  full_url <- .tce_build_request_url(endpoint, params)

  if (verbose) .tce_log_request(endpoint, full_url, params)

  req <- httr2::request(full_url) |>
    httr2::req_retry(max_tries = max_tries) |>
    httr2::req_user_agent("tceper R package (https://github.com/StrategicProjects/tceper)") |>
    httr2::req_error(is_error = function(resp) FALSE)

  if (progress) {
    cli::cli_progress_step(
      "Querying {.field {endpoint}} from the TCE-PE API\u2026",
      msg_done = "Query to {.field {endpoint}} completed.",
      msg_failed = "Failed to query {.field {endpoint}}."
    )
  }

  started <- Sys.time()
  resp <- withCallingHandlers(
    httr2::req_perform(req),
    error = function(e) .tce_explain_request_error(e, endpoint, full_url, call = .user_call)
  )
  elapsed <- as.numeric(difftime(Sys.time(), started, units = "secs"))

  if (verbose) .tce_log_response(resp, elapsed)

  .tce_check_response(resp, endpoint, full_url, call = .user_call)

  parsed <- tryCatch(
    jsonlite::fromJSON(httr2::resp_body_string(resp, encoding = "ISO-8859-1")),
    error = function(e) {
      cli::cli_abort(
        c(
          "x" = "The API response could not be parsed as JSON.",
          "i" = "Endpoint: {.field {endpoint}}",
          "i" = "URL: {.url {full_url}}",
          "i" = "Underlying error: {.emph {conditionMessage(e)}}",
          ">" = "Re-run with {.code verbose = TRUE} to see the raw response."
        ),
        call = .user_call
      )
    }
  )

  resposta <- parsed[["resposta"]]
  if (is.null(resposta)) {
    cli::cli_abort(
      c(
        "x" = "Unexpected response from the TCE-PE API.",
        "i" = "Endpoint {.field {endpoint}} did not return a {.field resposta} field.",
        "i" = "URL: {.url {full_url}}"
      ),
      call = .user_call
    )
  }

  status <- resposta[["status"]]
  if (!identical(status, "OK")) {
    mensagem <- resposta[["mensagem"]] %||% resposta[["message"]] %||% NA_character_
    cli::cli_abort(
      c(
        "x" = "The API returned status {.val {status}} for {.field {endpoint}}.",
        if (!is.na(mensagem)) c("i" = "Message: {.emph {mensagem}}"),
        "i" = "URL: {.url {full_url}}",
        ">" = "Check your query parameters with {.code tce_params(\"{endpoint}\")}."
      ),
      call = .user_call
    )
  }

  tamanho <- as.integer(resposta[["tamanhoResultado"]] %||% 0L)
  limite  <- as.integer(resposta[["limiteResultado"]] %||% 100000L)

  if (tamanho == 0L) {
    if (progress) {
      cli::cli_alert_warning(c(
        "No records found for {.field {endpoint}}.",
        if (length(params)) " Try relaxing or removing filters."
      ))
    }
    return(tibble::tibble())
  }

  if (progress) {
    cli::cli_alert_info("{.val {tamanho}} record{?s} returned in {.val {round(elapsed, 2)}}s.")
  }

  if (tamanho >= limite) {
    cli::cli_warn(c(
      "!" = "Result reached the API limit of {.val {limite}} records.",
      "i" = "Some rows are likely missing.",
      ">" = "Use additional filters (e.g. {.code anoreferencia}, {.code mesreferencia}) to narrow the query."
    ))
  }

  conteudo <- resposta[["conteudo"]]
  if (is.null(conteudo) || length(conteudo) == 0L) {
    return(tibble::tibble())
  }

  resultado <- tibble::as_tibble(conteudo)
  if (clean_names) resultado <- janitor::clean_names(resultado)
  resultado
}


# ---- Internal helpers --------------------------------------------------------

#' Build a parameter list, dropping NULLs
#' @noRd
tce_build_params <- function(...) {
  purrr::compact(list(...))
}

#' Build the actual request URL (preserves the literal '!')
#'
#' The TCE-PE API runs on a Latin-1 backend: query values containing
#' accented characters must be sent as ISO-8859-1 bytes, otherwise the
#' server matches no rows. We transcode UTF-8 inputs to Latin-1, mark
#' the result as raw bytes, then percent-encode.
#' @noRd
.tce_build_request_url <- function(endpoint, params) {
  url <- paste0("https://sistemas.tcepe.tc.br/DadosAbertos/", endpoint, "!json")
  if (length(params) == 0L) return(url)
  query <- paste(
    names(params),
    vapply(params, .tce_url_encode_latin1, character(1)),
    sep = "=", collapse = "&"
  )
  paste0(url, "?", query)
}

#' Percent-encode a query value using ISO-8859-1 byte semantics
#' @noRd
.tce_url_encode_latin1 <- function(value) {
  value <- as.character(value)
  if (!nzchar(value)) return(value)
  # Plain ASCII fast path
  if (!grepl("[^\x01-\x7f]", value, useBytes = TRUE)) {
    return(utils::URLencode(value, reserved = TRUE))
  }
  # Coerce to UTF-8 first so iconv has a defined input encoding,
  # then transcode to Latin-1. Lossy characters become "?".
  enc <- Encoding(value)
  if (enc != "UTF-8") value <- enc2utf8(value)
  latin1 <- iconv(value, from = "UTF-8", to = "ISO-8859-1", sub = "?")
  # Mark as raw bytes so URLencode operates byte-by-byte instead of
  # trying to decode the string as UTF-8.
  Encoding(latin1) <- "bytes"
  utils::URLencode(latin1, reserved = TRUE)
}

#' Build a human-readable URL string (with literal '!'). Validates and maps
#' parameters; kept for backward compatibility with callers that want a
#' pretty URL outside of `tce_request()`.
#' @noRd
tce_build_url <- function(endpoint, ..., call = rlang::caller_call()) {
  params <- tce_build_params(...)
  .tce_assert_allowed_params(endpoint, params, call = call)
  params <- .tce_map_params(endpoint, params)
  .tce_build_request_url(endpoint, params)
}

#' Build a cache key from endpoint and parameters
#' @noRd
tce_cache_key <- function(endpoint, ..., call = rlang::caller_call()) {
  params <- tce_build_params(...)
  .tce_assert_allowed_params(endpoint, params, call = call)
  params <- .tce_map_params(endpoint, params)
  if (length(params) == 0L) return(endpoint)
  params <- params[order(names(params))]
  paste0(endpoint, "?", paste(names(params), params, sep = "=", collapse = "&"))
}

#' Verbose logging -- request side
#' @noRd
.tce_log_request <- function(endpoint, url, params) {
  cli::cli_rule(left = "tceper request")
  cli::cli_alert_info("Endpoint: {.field {endpoint}}")
  cli::cli_alert_info("URL: {.url {url}}")
  if (length(params)) {
    cli::cli_alert_info("Parameters ({length(params)}):")
    cli::cli_ul(.tce_format_params(params))
  } else {
    cli::cli_alert_info("Parameters: {.emph (none)}")
  }
}

#' Verbose logging -- response side
#' @noRd
.tce_log_response <- function(resp, elapsed_secs) {
  status <- httr2::resp_status(resp)
  ctype  <- tryCatch(httr2::resp_content_type(resp), error = function(e) "unknown")
  bytes  <- tryCatch(length(httr2::resp_body_raw(resp)), error = function(e) NA_integer_)
  size   <- .tce_format_bytes(bytes)
  elapsed_s <- round(elapsed_secs, 2)

  cli::cli_alert_info(
    "HTTP {.val {status}} \u00b7 {ctype} \u00b7 {size} \u00b7 {elapsed_s}s"
  )
}

#' Format param list as 'name = value' bullets, truncating long values
#' @noRd
.tce_format_params <- function(params) {
  vals <- vapply(params, function(v) {
    s <- paste(as.character(v), collapse = ", ")
    if (nchar(s) > 60L) paste0(substr(s, 1L, 57L), "\u2026") else s
  }, character(1))
  stats::setNames(
    sprintf("{.field %s} = {.val %s}", names(params), vals),
    rep("*", length(params))
  )
}

#' Human-friendly byte count
#' @noRd
.tce_format_bytes <- function(n) {
  if (is.na(n)) return("? B")
  units <- c("B", "KB", "MB", "GB")
  i <- max(1L, min(length(units), floor(log(max(n, 1), 1024)) + 1L))
  sprintf("%.1f %s", n / 1024^(i - 1), units[i])
}

#' Check that an HTTP response is valid (status 2xx and JSON content)
#' @noRd
.tce_check_response <- function(resp, endpoint, url, call = NULL) {
  if (httr2::resp_is_error(resp)) {
    status <- httr2::resp_status(resp)
    desc   <- httr2::resp_status_desc(resp)
    body   <- .tce_safe_body_excerpt(resp)
    cli::cli_abort(
      c(
        "x" = "HTTP {.val {status}} {desc} when calling {.field {endpoint}}.",
        "i" = "URL: {.url {url}}",
        if (nzchar(body)) c("i" = "Response body: {.emph {body}}")
      ),
      call = call
    )
  }

  ctype <- tryCatch(httr2::resp_content_type(resp), error = function(e) "")
  if (!grepl("json|javascript", ctype, ignore.case = TRUE)) {
    body <- .tce_safe_body_excerpt(resp)
    cli::cli_abort(
      c(
        "x" = "The API did not return JSON for {.field {endpoint}}.",
        "i" = "Content-Type received: {.val {ctype}}",
        "i" = "URL: {.url {url}}",
        if (nzchar(body)) c("i" = "Response excerpt: {.emph {body}}")
      ),
      call = call
    )
  }
}

#' First ~200 chars of the response body, single-line, safely
#' @noRd
.tce_safe_body_excerpt <- function(resp, max_chars = 200L) {
  tryCatch({
    s <- httr2::resp_body_string(resp, encoding = "ISO-8859-1")
    s <- gsub("\\s+", " ", s)
    if (nchar(s) > max_chars) paste0(substr(s, 1L, max_chars - 1L), "\u2026") else s
  }, error = function(e) "")
}

#' Translate low-level libcurl/httr2 errors into actionable messages.
#' Covers: TLS handshake (LibreSSL key-exchange), DNS resolution, timeouts.
#' @noRd
.tce_explain_request_error <- function(cnd, endpoint, url, call = NULL) {
  msg <- paste(conditionMessage(cnd), collapse = " ")
  parent <- cnd$parent
  if (!is.null(parent)) msg <- paste(msg, paste(conditionMessage(parent), collapse = " "))

  is_tls <- grepl("SSL connect error|TLS connect error|certificate|handshake",
                  msg, ignore.case = TRUE)
  is_dns <- grepl("Could not resolve host|getaddrinfo|Name or service not known",
                  msg, ignore.case = TRUE)
  is_timeout <- grepl("Timeout|timed out|Operation too slow", msg, ignore.case = TRUE)

  if (is_dns) {
    cli::cli_abort(
      c(
        "x" = "Could not resolve {.url sistemas.tcepe.tc.br}.",
        "i" = "The TCE-PE API only accepts connections from Brazilian IP addresses.",
        ">" = "Check your internet connection, DNS settings, or VPN location.",
        "i" = "Endpoint: {.field {endpoint}}"
      ),
      call = call, parent = cnd
    )
  }

  if (is_tls) {
    cli::cli_abort(
      c(
        "x" = "TLS handshake with {.url sistemas.tcepe.tc.br} failed.",
        "i" = "If you are outside Brazil, the API is geo-restricted and may drop the connection at the TLS layer.",
        ">" = "Verify that your network can reach the host from a Brazilian IP."
      ),
      call = call, parent = cnd
    )
  }

  if (is_timeout) {
    cli::cli_abort(
      c(
        "x" = "Request to {.field {endpoint}} timed out.",
        "i" = "The TCE-PE API can be slow for broad queries.",
        ">" = "Add more filters to narrow the query, or increase {.code max_tries}."
      ),
      call = call, parent = cnd
    )
  }

  # Fall through: let httr2's error propagate
  invisible()
}

# ---- Public works ------------------------------------------------------------

#' Public works
#'
#' Retrieve engineering works and services overseen by the Court of Accounts.
#'
#' @param ... Named query parameters passed directly to the API.
#'   Common filters: `CodigoEfiscoUG`, `CodigoMunicipio`.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family public_works
tce_public_works <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Obras", ..., call = .user_call), tce_request("Obras", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Public works contractor data
#'
#' Retrieve data on companies contracted for public works.
#'
#' @inheritParams tce_public_works
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family public_works
tce_public_works_contractors <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("ObrasDadosContratacao", ..., call = .user_call), tce_request("ObrasDadosContratacao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Public works audit data
#'
#' Retrieve audit data on public works.
#'
#' @inheritParams tce_public_works
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family public_works
tce_public_works_audits <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("DadosObrasAuditoria", ..., call = .user_call), tce_request("DadosObrasAuditoria", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' School transport
#'
#' Retrieve school transport services overseen by the Court of Accounts.
#'
#' @inheritParams tce_public_works
#' @return A [tibble][tibble::tibble-package].
#' @export
tce_school_transport <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TransporteEscolar", ..., call = .user_call), tce_request("TransporteEscolar", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- Personnel ---------------------------------------------------------------

#' Public servants
#'
#' Retrieve municipal and state public servants (SAGRES Personnel).
#'
#' @param ... Named query parameters passed directly to the API.
#'   Common filters: `CodigoEfiscoUG`, `Nome`, `Cpf`, `CodigoMunicipio`.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family personnel
tce_servants <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("ListaServidores", ..., call = .user_call), tce_request("ListaServidores", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- Reference tables --------------------------------------------------------

#' Municipalities
#'
#' Retrieve the list of municipalities in Pernambuco.
#'
#' @param ... Named query parameters passed directly to the API.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family reference
tce_municipalities <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("Municipios", ..., call = .user_call), tce_request("Municipios", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Managed entities (all)
#'
#' Retrieve all managed entities (*unidades jurisdicionadas*), state
#' and municipal.
#'
#' @param ... Named query parameters passed directly to the API.
#'   Common filters: `Esfera` (`"E"` state, `"M"` municipal), `Natureza`.
#' @param verbose If `TRUE`, prints the final API URL.
#' @param cache If `TRUE` (default), returns cached results when available.
#' @param clean_names If `TRUE` (default), converts column names to `snake_case`.
#'
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family reference
tce_entities <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("UnidadesJurisdicionadas", ..., call = .user_call), tce_request("UnidadesJurisdicionadas", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' State managed entities
#'
#' Retrieve state-level managed entities.
#'
#' @inheritParams tce_municipalities
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family reference
tce_state_entities <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("UnidadesJurisdicionadasEstaduais", ..., call = .user_call), tce_request("UnidadesJurisdicionadasEstaduais", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Municipal managed entities
#'
#' Retrieve municipal-level managed entities.
#'
#' @param ... Named query parameters. Common: `CodigoMunicipio`.
#' @inheritParams tce_municipalities
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family reference
tce_municipal_entities <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("UnidadesJurisdicionadasMunicipais", ..., call = .user_call), tce_request("UnidadesJurisdicionadasMunicipais", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Sub-units of managed entities
#'
#' Retrieve sub-units belonging to managed entities.
#'
#' @param ... Named query parameters. Common: `CodigoEfiscoUG`.
#' @inheritParams tce_municipalities
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family reference
tce_sub_units <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("SubunidadesUnidadesJurisdicionadas", ..., call = .user_call), tce_request("SubunidadesUnidadesJurisdicionadas", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Creditor types
#'
#' Lookup table of supplier/creditor types.
#'
#' @inheritParams tce_municipalities
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family reference
tce_creditor_types <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TipoCredor", ..., call = .user_call), tce_request("TipoCredor", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Payroll types
#'
#' Lookup table of payroll (*folha*) types.
#'
#' @inheritParams tce_municipalities
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family reference
tce_payroll_types <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TipoFolha", ..., call = .user_call), tce_request("TipoFolha", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Funding source types
#'
#' Lookup table of funding source types.
#'
#' @inheritParams tce_municipalities
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family reference
tce_funding_sources <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TipoFonteRecurso", ..., call = .user_call), tce_request("TipoFonteRecurso", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

#' Inactivation reason types
#'
#' Lookup table of inactivation reason types.
#'
#' @inheritParams tce_municipalities
#' @return A [tibble][tibble::tibble-package].
#' @export
#' @family reference
tce_inactivation_reasons <- function(..., verbose = getOption("tceper.verbose", FALSE), cache = TRUE, clean_names = TRUE) {
  .user_call <- rlang::caller_call()
  tce_cached(tce_cache_key("TipoMotivoInativacao", ..., call = .user_call), tce_request("TipoMotivoInativacao", ..., clean_names = clean_names, verbose = verbose), use_cache = cache)
}

# ---- Endpoint catalog & parameter mapping ------------------------------------

`%||%` <- function(x, y) if (is.null(x)) y else x

.tce_catalog_env <- new.env(parent = emptyenv())

.tce_catalog <- function() {
  if (exists("data", envir = .tce_catalog_env, inherits = FALSE)) {
    return(get("data", envir = .tce_catalog_env, inherits = FALSE))
  }

  path <- system.file("extdata", "tcepe_dados_abertos_catalogo.json", package = "tceper")
  if (!nzchar(path) || !file.exists(path)) {
    assign("data", NULL, envir = .tce_catalog_env)
    return(NULL)
  }

  # Keep nested structures as lists
  cat <- jsonlite::fromJSON(path, simplifyVector = FALSE)

  # Defensive: ensure expected shape
  if (is.null(cat$metodos)) {
    assign("data", NULL, envir = .tce_catalog_env)
    return(NULL)
  }

  assign("data", cat, envir = .tce_catalog_env)
  cat
}

.tce_find_method <- function(endpoint, catalog = .tce_catalog()) {
  if (is.null(catalog)) return(NULL)
  methods <- catalog$metodos
  if (!is.list(methods)) return(NULL)

  ep_norm <- tolower(endpoint)

  # metodos is a list of methods (simplifyVector=FALSE)
  for (m in methods) {
    ent <- m$entidade %||% m$endpoint %||% ""
    if (tolower(ent) == ep_norm) return(m)
  }

  # Fallback: try removing underscores/spaces and comparing
  ep_key <- gsub("[^a-z0-9]+", "", ep_norm)
  for (m in methods) {
    ent <- m$entidade %||% m$endpoint %||% ""
    ent_key <- gsub("[^a-z0-9]+", "", tolower(ent))
    if (ent_key == ep_key) return(m)
  }

  NULL
}

.tce_param_key <- function(x) {
  gsub("[^a-z0-9]+", "", tolower(x %||% ""))
}

.tce_build_param_lookup <- function(method) {
  entrada <- method$entrada
  if (is.null(entrada) || length(entrada) == 0) return(list())

  # entrada is expected as list of param objects
  lookup <- list()
  for (p in entrada) {
    api_name <- p$api_name %||% p$nome %||% p$param %||% NULL
    if (is.null(api_name)) next

    # Match against API name and common variants
    keys <- unique(c(
      .tce_param_key(api_name),
      .tce_param_key(gsub("_", "", api_name)),
      .tce_param_key(tolower(api_name))
    ))

    # Also accept snake_case user inputs for CamelCase api names
    snake <- janitor::make_clean_names(api_name)
    keys <- unique(c(keys, .tce_param_key(snake)))

    for (k in keys) {
      lookup[[k]] <- api_name
    }
  }

  lookup
}

# Map user-provided parameter names to the API's official api_name
.tce_map_params <- function(endpoint, params) {
  if (length(params) == 0L) return(params)

  method <- .tce_find_method(endpoint)
  if (is.null(method)) return(params)

  lookup <- .tce_build_param_lookup(method)
  if (length(lookup) == 0L) return(params)

  new_names <- names(params)
  for (i in seq_along(new_names)) {
    nm <- new_names[[i]]
    key <- .tce_param_key(nm)
    api <- lookup[[key]]
    if (!is.null(api) && nzchar(api)) {
      new_names[[i]] <- api
    }
  }
  names(params) <- new_names
  params
}


# Validate that user-provided query parameters exist for the given endpoint.
# If unknown parameters are found, abort with a helpful cli list of allowed
# parameters. The list displays the R-friendly `r_name` as {.field ...} and the
# exact API parameter name (`api_name`) in plain text.
.tce_allowed_params_tbl <- function(method) {
  entrada <- method$entrada %||% list()
  if (is.null(entrada) || length(entrada) == 0L) {
    return(tibble::tibble(r_name = character(), api_name = character()))
  }

  api <- vapply(
    entrada,
    function(p) p$api_name %||% p$nome %||% p$param %||% p$parametro %||% NA_character_,
    character(1)
  )
  api <- api[!is.na(api) & nzchar(api)]
  if (length(api) == 0L) {
    return(tibble::tibble(r_name = character(), api_name = character()))
  }

  tibble::tibble(
    r_name   = janitor::make_clean_names(api),
    api_name = api
  )
}

.tce_abort_unknown_params <- function(endpoint, unknown, allowed_tbl, call = rlang::caller_env()) {
  unknown_str <- paste(unknown, collapse = ", ")

  if (!requireNamespace("cli", quietly = TRUE)) {
    allowed_str <- paste(sprintf("%s (%s)", allowed_tbl$r_name, allowed_tbl$api_name), collapse = ", ")
    stop(
      sprintf(
        "Unknown query parameter(s) for endpoint %s: %s. Allowed parameters: %s",
        endpoint, unknown_str, allowed_str
      ),
      call. = FALSE
    )
  }

  cli::cli_alert_danger("Unknown query parameter(s) for endpoint {.field {endpoint}}: {unknown_str}")
  cli::cli_alert_info("Allowed parameters:")
  cli::cli_ul()
  for (i in seq_len(nrow(allowed_tbl))) {
    invisible(cli::cli_li("{.field {allowed_tbl$r_name[i]}} ({allowed_tbl$api_name[i]})"))
  }
  cli::cli_abort("Invalid query parameter(s).", call = call)
}

.tce_assert_allowed_params <- function(endpoint, params, call = rlang::caller_env()) {
  if (length(params) == 0L) return(invisible(TRUE))

  method <- .tce_find_method(endpoint)
  if (is.null(method)) return(invisible(TRUE))

  lookup <- .tce_build_param_lookup(method)
  if (length(lookup) == 0L) return(invisible(TRUE))

  nm <- names(params)
  if (is.null(nm) || !length(nm)) return(invisible(TRUE))

  unknown <- nm[!vapply(nm, function(x) !is.null(lookup[[.tce_param_key(x)]]), logical(1))]
  if (length(unknown)) {
    allowed_tbl <- .tce_allowed_params_tbl(method)
    # Sort for stable display
    if (nrow(allowed_tbl)) {
      allowed_tbl <- allowed_tbl[order(allowed_tbl$r_name, allowed_tbl$api_name), , drop = FALSE]
    }
    .tce_abort_unknown_params(endpoint, unknown, allowed_tbl, call = call)
  }

  invisible(TRUE)
}

---
title: "Exploring endpoints, parameters, and fields"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Exploring endpoints, parameters, and fields}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE,
  purl = FALSE  # The TCE-PE API is geo-restricted (Brazil-only). Skip code
                # extraction so R CMD check does not try to run it on CRAN
                # machines outside Brazil.
)
```

> **Note** — Code chunks are shown but not executed at build time
> (`eval = FALSE`). The discovery snippets (`tce_catalog()`,
> `tce_params()`, `tce_fields()`) work offline anywhere and can be run
> as-is from any network. The query snippets require a Brazilian IP
> (the TCE-PE API is geo-restricted).

This vignette shows how to *discover* endpoints and their input/output
metadata using the built-in catalog shipped with **tceper**.

## Discover endpoints with `tce_catalog()`

```{r catalog}
library(tceper)

tce_catalog()
```

Filter by keyword:

```{r catalog-search}
tce_catalog(search = "licit")
```

## Inspect one endpoint

Pick an endpoint name from `tce_catalog()`. For instance, `LicitacaoUG`:

### Input parameters

```{r params}
tce_params("LicitacaoUG")
```

`api_name` is the exact query name expected by the API.

`r_name` is the `snake_case` name you can pass to the R wrapper.

### Output fields

```{r fields}
tce_fields("LicitacaoUG")
```

## Query using `snake_case` arguments

Once you know which parameters exist, you can query with `snake_case`
arguments (the package maps them to `api_name` internally).

```{r query}
tce_bids(anomodalidade = "2025")
```

## Parameter validation

`tceper` validates query parameters against the endpoint catalog. If you pass
a parameter that does not exist for that endpoint, it aborts and shows the
allowed parameter list.

```{r invalid, eval = FALSE}
# Example (invalid parameter name for this endpoint):
tce_contracts(fog = "510101", verbose = TRUE)
```

To avoid this, inspect allowed parameters first:

```{r inspect-first}
tce_params("Contratos")
```

## Working with cache

Many wrapper functions can cache results in-memory for faster repeated calls.

You can inspect cache state with:

```{r cache-info}
tce_cache_info()
```

And clear it with:

```{r cache-clear}
tce_cache_clear()
```

When `verbose = TRUE`, cache hits may print a short message indicating the
returned object comes from cache.

---
title: "API endpoints reference"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{API endpoints reference}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This vignette provides a *generated* mapping between exported `tceper`
functions and their underlying TCE-PE API endpoints, based on the package
code and the built-in catalog.

The table below is generated locally **from the built-in catalog** — no
API call is made, so it builds anywhere regardless of network access.
The example block at the end of the vignette is shown but not executed
(the API is only reachable from Brazilian IP addresses).

```{r mapping, eval = TRUE}
library(tceper)

# Catalog with endpoint metadata
catalog <- tce_catalog()

# Discover exported wrappers and extract the endpoint used internally.
# We look for calls of the form: tce_cached("Endpoint") or tce_request("Endpoint")
exports <- getNamespaceExports("tceper")

# Keep likely "wrapper" functions
skip <- c(
  "tce_request", "tce_catalog", "tce_endpoint", "tce_params", "tce_fields",
  "tce_cache_info", "tce_cache_clear"
)
wrappers <- setdiff(exports, skip)

extract_endpoint <- function(fn) {
  f <- get(fn, envir = asNamespace("tceper"))
  txt <- paste(deparse(body(f)), collapse = "\n")
  m <- regexpr('tce_(?:cached|request)\\("([^"]+)"', txt, perl = TRUE)
  if (m[1] == -1) return(NA_character_)
  hit <- regmatches(txt, m)
  sub('^tce_(?:cached|request)\\("([^"]+)".*$', "\\1", hit)
}

endpoint <- vapply(wrappers, extract_endpoint, character(1))

mappings <- tibble::tibble(
  r_function = paste0("`", wrappers, "()`"),
  endpoint   = endpoint
) |>
  dplyr::filter(!is.na(endpoint)) |>
  dplyr::left_join(
    catalog |> dplyr::select(endpoint, group, title),
    by = "endpoint"
  ) |>
  dplyr::arrange(group, endpoint, r_function)

mappings
```

## How to use this reference

- Use `tce_catalog(search = "...")` to locate an endpoint of interest.
- Use `tce_params("Endpoint")` to discover available query parameters.
- Use `tce_fields("Endpoint")` to see the output schema.
- Then call the corresponding wrapper function (as listed above) or call
  `tce_request("Endpoint", ...)` directly with API parameter names.

### Example: inspect and query `Contratos`

```{r example, eval = FALSE, purl = FALSE}
tce_params("Contratos")
tce_fields("Contratos")

tce_contracts(codigo_efisco_ug = "510101", ano_contrato = "2025")
```

---
title: "Getting started with tceper"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with tceper}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE,
  purl = FALSE  # The TCE-PE API is geo-restricted (Brazil-only). Skip code
                # extraction so R CMD check does not try to run it on CRAN
                # machines outside Brazil.
)
```

> **Note** — Code chunks in this vignette are shown but not executed
> at build time (`eval = FALSE`). The TCE-PE API host
> (`sistemas.tcepe.tc.br`) only accepts connections from Brazilian IP
> addresses, so requests would fail on CRAN check machines and on most
> CI runners outside Brazil. Run the snippets interactively from a
> Brazilian network to reproduce the outputs. The discovery functions
> (`tce_catalog()`, `tce_params()`, `tce_fields()`) work offline
> anywhere.

## Overview

**tceper** is an R client for the TCE-PE (Tribunal de Contas do Estado de Pernambuco) Open Data API.

The package wraps the official endpoints into user-friendly functions that:

- accept `snake_case` arguments (mapped to the API's expected query names), and
- return tibbles (with optional `snake_case` output column names).

## Installation

```{r install}
# install.packages("remotes")
remotes::install_github("StrategicProjects/tceper")
```

## Load the package

```{r setup}
library(tceper)
```

## Explore endpoints before querying

### List endpoints

```{r catalog}
tce_catalog()
```

Filter by keyword:

```{r catalog-search}
tce_catalog(search = "contrat")
```

### Inspect input parameters

```{r params}
tce_params("Contratos")
```

### Inspect output fields

```{r fields}
tce_fields("Contratos")
```

## Quick queries

### Municipalities

The `Municipios` endpoint is exposed as `tce_municipalities()`.

```{r municipalities}
tce_municipalities()
```

You can filter by state (UF) or by municipality name. Accented
Portuguese characters work as-is -- the package transcodes UTF-8 input
to the Latin-1 encoding the API expects:

```{r municipalities-filter}
tce_municipalities(unidadefederativa = "PE")
tce_municipalities(municipio = "Recife")
tce_municipalities(municipio = "São José da Coroa Grande")
```

### Contracts

```{r contracts}
tce_contracts(codigo_efisco_ug = "510101")
```

You can add additional filters. Use `tce_params("Contratos")` to see what is allowed.

```{r contracts-filter}
tce_contracts(codigo_efisco_ug = "510101", ano_contrato = "2025")
```

### Suppliers

The `Fornecedores` endpoint is exposed as `tce_suppliers()`.

```{r suppliers}
tce_suppliers(cpfcnpj = "25173110000186")
tce_suppliers(nome = "eireli")
```

## Verbose mode (recommended while exploring)

When `verbose = TRUE`, the package prints:

- the final API URL, and
- helper commands to inspect inputs/outputs for the endpoint.

```{r verbose}
tce_bids(anomodalidade = "2025", verbose = TRUE)
```

## Keeping original output field names

By default, output columns are converted to `snake_case`. To keep the original
names from the API:

```{r clean-names}
tce_municipalities(clean_names = FALSE)
```

## Handling the 100,000-record limit

The API returns at most **100,000 records** per request. When this limit is
reached, `tceper` issues a warning.

To work around this, apply filters to narrow your query:

```{r limits}
# Avoid unfiltered requests that may hit the 100k limit:
# tce_municipal_expenditures()

# Choose a municipality code:
library(dplyr)
library(stringr)

tce_municipalities() |>
  filter(str_detect(municipio, "Pesqueira")) |>
  glimpse()
# Rows: 1
# Columns: 5
# $ codigoibge        <chr> "2610905"
# $ codigo            <chr> "P113"
# $ unidadefederativa <chr> "PE"
# $ municipio         <chr> "Pesqueira"
# $ codigosagres      <chr> "5408"

# Filter by municipality and year:
tce_municipal_expenditures(
  codigo_municipio = "P113", # Pesqueira
  ano_referencia   = "2025"
)
```

## Direct access to any endpoint

If you prefer to call an endpoint directly, use `tce_request()` and pass the
API query names (as shown by `tce_params()`):

```{r request}
tce_request("Contratos", CodigoEfiscoUG = "510101", AnoContrato = "2025")
```

## Cache

All wrapper functions cache results in memory. The default TTL is 1 hour.

```{r cache}
tce_contracts(codigo_efisco_ug = "510101")  # hits the API
tce_contracts(codigo_efisco_ug = "510101")  # instant (from cache)

# Bypass cache
tce_contracts(codigo_efisco_ug = "510101", cache = FALSE)

# Inspect and manage
tce_cache_info()
tce_cache_clear()
```

You can change the TTL globally:

```{r cache-options}
options(tceper.cache_ttl = 7200)  # 2 hours
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/introspection.R
\name{tce_params}
\alias{tce_params}
\title{List input parameters for an endpoint}
\usage{
tce_params(endpoint)
}
\arguments{
\item{endpoint}{Endpoint name (e.g., \code{"Municipios"}, \code{"Contratos"}).
Use \code{\link[=tce_catalog]{tce_catalog()}} to see all available endpoints.}
}
\value{
A tibble with columns \code{api_name}, \code{r_name}, \code{required}, \code{type},
and \code{description}. Has a custom print method that summarises the
parameter names at the top.
}
\description{
Returns a tibble describing the query parameters accepted by a given
endpoint. The \code{api_name} column shows the exact name expected by the
API; the \code{r_name} column shows the \code{snake_case} equivalent that you
can pass to wrapper functions.
}
\examples{
tce_params("Contratos")
tce_params("Municipios")
tce_params("LicitacaoUG")
}
\seealso{
\code{\link[=tce_fields]{tce_fields()}}, \code{\link[=tce_catalog]{tce_catalog()}}, \code{\link[=tce_endpoint]{tce_endpoint()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{tce_cache_clear}
\alias{tce_cache_clear}
\title{Clear the tceper in-memory cache}
\usage{
tce_cache_clear(pattern = NULL)
}
\arguments{
\item{pattern}{Optional regex pattern to clear only matching keys.
If \code{NULL} (default), clears the entire cache. For example,
\code{"Contratos"} clears all cached contract queries.}
}
\value{
Invisibly returns the number of entries cleared.
}
\description{
Removes cached API responses so that subsequent calls hit the API again.
By default clears everything; use \code{pattern} to selectively clear entries
for specific endpoints.
}
\examples{
tce_cache_clear()              # clear everything
tce_cache_clear("Contratos")   # clear contract queries only
}
\seealso{
\code{\link[=tce_cache_info]{tce_cache_info()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/processes.R
\name{tce_debts_fines}
\alias{tce_debts_fines}
\title{Debts and fines}
\usage{
tce_debts_fines(
  ...,
  verbose = getOption("tceper.verbose", FALSE),
  cache = TRUE,
  clean_names = TRUE
)
}
\arguments{
\item{...}{Named query parameters passed directly to the API.
Common filters: \code{CodigoEfiscoUG}, \code{AnoReferencia}, \code{CodigoMunicipio}.}

\item{verbose}{If \code{TRUE}, prints the final API URL.}

\item{cache}{If \code{TRUE} (default), returns cached results when available.}

\item{clean_names}{If \code{TRUE} (default), converts column names to
\code{snake_case}.}
}
\value{
A \link[tibble:tibble-package]{tibble}.
}
\description{
Retrieve debts and fines imposed.
}
\seealso{
Other processes:
\code{\link[=tce_consideranda]{tce_consideranda()}},
\code{\link[=tce_determinations]{tce_determinations()}},
\code{\link[=tce_outcomes]{tce_outcomes()}},
\code{\link[=tce_processes]{tce_processes()}},
\code{\link[=tce_recommendations]{tce_recommendations()}},
\code{\link[=tce_retirement_outcomes]{tce_retirement_outcomes()}},
\code{\link[=tce_special_accountability]{tce_special_accountability()}},
\code{\link[=tce_spending_limits]{tce_spending_limits()}}
}
\concept{processes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revenues.R
\name{tce_municipal_revenues}
\alias{tce_municipal_revenues}
\title{Municipal revenues}
\usage{
tce_municipal_revenues(
  ...,
  verbose = getOption("tceper.verbose", FALSE),
  cache = TRUE,
  clean_names = TRUE
)
}
\arguments{
\item{...}{Named query parameters passed directly to the API.
Common filters: \code{CodigoEfiscoUG}, \code{AnoReferencia}.}

\item{verbose}{If \code{TRUE}, prints the final API URL.}

\item{cache}{If \code{TRUE} (default), returns cached results when available.}

\item{clean_names}{If \code{TRUE} (default), converts column names to
\code{snake_case}.}
}
\value{
A \link[tibble:tibble-package]{tibble}.
}
\description{
Retrieve municipal-level revenue data.
}
\examples{
\dontrun{
# Not run: reaches the live TCE-PE API, which only accepts connections
# from Brazilian IP addresses (unreachable from CRAN / most CI).
tce_municipal_revenues(id_unidade_gestora = "384")
}
}
\seealso{
Other revenues:
\code{\link[=tce_budgeted_revenues]{tce_budgeted_revenues()}},
\code{\link[=tce_state_revenues]{tce_state_revenues()}}
}
\concept{revenues}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reference.R
\name{tce_state_entities}
\alias{tce_state_entities}
\title{State managed entities}
\usage{
tce_state_entities(
  ...,
  verbose = getOption("tceper.verbose", FALSE),
  cache = TRUE,
  clean_names = TRUE
)
}
\arguments{
\item{...}{Named query parameters passed directly to the API.}

\item{verbose}{If \code{TRUE}, prints the final API URL.}

\item{cache}{If \code{TRUE} (default), returns cached results when available.}

\item{clean_names}{If \code{TRUE} (default), converts column names to \code{snake_case}.}
}
\value{
A \link[tibble:tibble-package]{tibble}.
}
\description{
Retrieve state-level managed entities.
}
\seealso{
Other reference:
\code{\link[=tce_creditor_types]{tce_creditor_types()}},
\code{\link[=tce_entities]{tce_entities()}},
\code{\link[=tce_funding_sources]{tce_funding_sources()}},
\code{\link[=tce_inactivation_reasons]{tce_inactivation_reasons()}},
\code{\link[=tce_municipal_entities]{tce_municipal_entities()}},
\code{\link[=tce_municipalities]{tce_municipalities()}},
\code{\link[=tce_payroll_types]{tce_payroll_types()}},
\code{\link[=tce_sub_units]{tce_sub_units()}}
}
\concept{reference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/procurement.R
\name{tce_bid_modalities}
\alias{tce_bid_modalities}
\title{Bid modalities}
\usage{
tce_bid_modalities(
  ...,
  verbose = getOption("tceper.verbose", FALSE),
  cache = TRUE,
  clean_names = TRUE
)
}
\arguments{
\item{...}{Named query parameters passed directly to the API.
Common filters: \code{CodigoEfiscoUG}, \code{AnoReferencia}, \code{CodigoMunicipio}.}

\item{verbose}{If \code{TRUE}, prints the final API URL.}

\item{cache}{If \code{TRUE} (default), returns cached results when available.}

\item{clean_names}{If \code{TRUE} (default), converts column names to
\code{snake_case}.}
}
\value{
A \link[tibble:tibble-package]{tibble}.
}
\description{
Lookup table of bid modality types.
}
\seealso{
Other procurement:
\code{\link[=tce_agreements]{tce_agreements()}},
\code{\link[=tce_bid_details]{tce_bid_details()}},
\code{\link[=tce_bid_documents]{tce_bid_documents()}},
\code{\link[=tce_bid_legal_basis]{tce_bid_legal_basis()}},
\code{\link[=tce_bid_stages]{tce_bid_stages()}},
\code{\link[=tce_bid_statuses]{tce_bid_statuses()}},
\code{\link[=tce_bids]{tce_bids()}},
\code{\link[=tce_contract_amendments]{tce_contract_amendments()}},
\code{\link[=tce_contract_documents]{tce_contract_documents()}},
\code{\link[=tce_contract_items]{tce_contract_items()}},
\code{\link[=tce_contracts]{tce_contracts()}},
\code{\link[=tce_object_characteristics]{tce_object_characteristics()}},
\code{\link[=tce_object_classifications]{tce_object_classifications()}},
\code{\link[=tce_object_nature]{tce_object_nature()}}
}
\concept{procurement}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/introspection.R
\name{tce_catalog}
\alias{tce_catalog}
\title{List available API endpoints}
\usage{
tce_catalog(search = NULL)
}
\arguments{
\item{search}{Optional character string to filter endpoints by name or
description (case-insensitive). For example, \code{search = "licit"} returns
all procurement-related endpoints.}
}
\value{
A tibble with columns \code{endpoint}, \code{group}, \code{title}, and \code{url}.
}
\description{
Returns a tibble listing all 71 endpoints available in the TCE-PE Open
Data API, with their group, description and URL. Use \code{search} to filter
by keyword.
}
\examples{
tce_catalog()
tce_catalog(search = "contrat")
tce_catalog(search = "municipio")
}
\seealso{
\code{\link[=tce_params]{tce_params()}}, \code{\link[=tce_fields]{tce_fields()}}, \code{\link[=tce_endpoint]{tce_endpoint()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/introspection.R
\name{tce_fields}
\alias{tce_fields}
\title{List output fields for an endpoint}
\usage{
tce_fields(endpoint)
}
\arguments{
\item{endpoint}{Endpoint name (e.g., \code{"Municipios"}, \code{"Contratos"}).
Use \code{\link[=tce_catalog]{tce_catalog()}} to see all available endpoints.}
}
\value{
A tibble with columns \code{name}, \code{type}, and \code{description}. Has a
custom print method that summarises the field names at the top.
}
\description{
Returns a tibble describing the columns returned by a given endpoint.
When \code{clean_names = TRUE} (the default), column names are converted
from the API's original names to \code{snake_case}.
}
\examples{
tce_fields("Contratos")
tce_fields("Municipios")
}
\seealso{
\code{\link[=tce_params]{tce_params()}}, \code{\link[=tce_catalog]{tce_catalog()}}, \code{\link[=tce_endpoint]{tce_endpoint()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tceper-package.R
\docType{package}
\name{tceper-package}
\alias{tceper}
\alias{tceper-package}
\title{tceper: Access the Open Data API of TCE-PE}
\description{
An R interface to the Open Data API of the Tribunal de Contas do Estado
de Pernambuco (TCE-PE), the Court of Accounts of the State of Pernambuco,
Brazil. Provides tidy, ready-to-use functions to query public data on
revenues, expenditures, commitments, procurement, contracts, agreements,
public works, legal processes, personnel and reference tables for all
state and municipal government entities in Pernambuco.
}
\section{Access to the API}{


The TCE-PE API host (\code{sistemas.tcepe.tc.br}) is \strong{geo-restricted}
and only accepts requests from Brazilian IP addresses. Users outside
Brazil will see connection or TLS errors. The discovery functions
(\code{\link[=tce_catalog]{tce_catalog()}}, \code{\link[=tce_endpoint]{tce_endpoint()}}, \code{\link[=tce_params]{tce_params()}}, \code{\link[=tce_fields]{tce_fields()}})
are unaffected -- they read from a built-in catalog and work offline
from any location.
}

\section{Recommended workflow}{

\enumerate{
\item \strong{Discover} endpoints with \code{\link[=tce_catalog]{tce_catalog()}} (filter with \verb{search =}).
\item \strong{Inspect} input parameters with \code{\link[=tce_params]{tce_params()}} and output fields
with \code{\link[=tce_fields]{tce_fields()}}.
\item \strong{Query} using a wrapper function (e.g. \code{\link[=tce_contracts]{tce_contracts()}}) or the
low-level \code{\link[=tce_request]{tce_request()}}.
}
}

\section{Parameter names}{


All query parameters are passed through \code{...} using the names shown
by \code{\link[=tce_params]{tce_params()}}. You can use either the original API name (e.g.
\code{CodigoEfiscoUG}) or its \code{snake_case} equivalent (e.g.
\code{codigo_efisco_ug}); the package maps them automatically.
}

\section{Options}{


\describe{
\item{\code{tceper.verbose}}{If \code{TRUE}, all functions print the final API
URL. Default: \code{FALSE}.}
\item{\code{tceper.progress}}{If \code{FALSE}, suppresses progress messages.
Default: \code{TRUE}.}
\item{\code{tceper.cache_ttl}}{Cache time-to-live in seconds.
Default: \code{3600} (1 hour).}
}
}

\seealso{
\itemize{
\item \code{\link[=tce_catalog]{tce_catalog()}}, \code{\link[=tce_params]{tce_params()}}, \code{\link[=tce_fields]{tce_fields()}} — endpoint discovery
\item \code{\link[=tce_request]{tce_request()}} — low-level query function
\item \code{\link[=tce_cache_info]{tce_cache_info()}}, \code{\link[=tce_cache_clear]{tce_cache_clear()}} — cache management
}
}
\author{
\strong{Maintainer}: Andre Leite \email{leite@castlab.org}

Authors:
\itemize{
  \item Andre Leite \email{leite@castlab.org}
  \item Marcos Wasilew \email{marcos.wasilew@gmail.com}
  \item Hugo Vasconcelos \email{hugo.vasconcelos@ufpe.br}
  \item Carlos Amorin \email{carlos.agaf@ufpe.br}
  \item Diogo Bezerra \email{diogo.bezerra@ufpe.br}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remessa.R
\name{tce_remessa_works}
\alias{tce_remessa_works}
\title{Public works details (Remessa)}
\usage{
tce_remessa_works(
  ...,
  verbose = getOption("tceper.verbose", FALSE),
  cache = TRUE,
  clean_names = TRUE
)
}
\arguments{
\item{...}{Named query parameters. Common: \code{CodigoEfiscoUG}.}

\item{verbose}{If \code{TRUE}, prints the final API URL.}

\item{cache}{If \code{TRUE} (default), returns cached results when available.}

\item{clean_names}{If \code{TRUE} (default), converts column names to \code{snake_case}.}
}
\value{
A \link[tibble:tibble-package]{tibble}.
}
\description{
Public works details (Remessa)
}
\seealso{
Other remessa:
\code{\link[=tce_benefit_types]{tce_benefit_types()}},
\code{\link[=tce_budget_statuses]{tce_budget_statuses()}},
\code{\link[=tce_legal_instrument_documents]{tce_legal_instrument_documents()}},
\code{\link[=tce_legal_instrument_items]{tce_legal_instrument_items()}},
\code{\link[=tce_legal_instrument_participants]{tce_legal_instrument_participants()}},
\code{\link[=tce_legal_instruments]{tce_legal_instruments()}},
\code{\link[=tce_procurement_process_budget]{tce_procurement_process_budget()}},
\code{\link[=tce_procurement_process_documents]{tce_procurement_process_documents()}},
\code{\link[=tce_procurement_process_participants]{tce_procurement_process_participants()}},
\code{\link[=tce_procurement_processes]{tce_procurement_processes()}},
\code{\link[=tce_reference_codes]{tce_reference_codes()}},
\code{\link[=tce_reference_dates]{tce_reference_dates()}},
\code{\link[=tce_reference_sources]{tce_reference_sources()}},
\code{\link[=tce_remessa_works_execution]{tce_remessa_works_execution()}},
\code{\link[=tce_remessa_works_geometry]{tce_remessa_works_geometry()}},
\code{\link[=tce_update_indices]{tce_update_indices()}}
}
\concept{remessa}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{tce_cache_info}
\alias{tce_cache_info}
\title{Show cache status}
\usage{
tce_cache_info()
}
\value{
A \link[tibble:tibble-package]{tibble} with columns \code{key}, \code{rows},
\code{cached_at} and \code{age_secs}, or an empty tibble if the cache is empty.
The cache TTL is controlled by \code{getOption("tceper.cache_ttl", 3600)}.
}
\description{
Returns a tibble describing all currently cached entries, including
the cache key (endpoint + parameters), row count, timestamp and age.
Useful for understanding what is cached and when it will expire.
}
\examples{
tce_cache_info()
}
\seealso{
\code{\link[=tce_cache_clear]{tce_cache_clear()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/public_works.R
\name{tce_school_transport}
\alias{tce_school_transport}
\title{School transport}
\usage{
tce_school_transport(
  ...,
  verbose = getOption("tceper.verbose", FALSE),
  cache = TRUE,
  clean_names = TRUE
)
}
\arguments{
\item{...}{Named query parameters passed directly to the API.
Common filters: \code{CodigoEfiscoUG}, \code{CodigoMunicipio}.}

\item{verbose}{If \code{TRUE}, prints the final API URL.}

\item{cache}{If \code{TRUE} (default), returns cached results when available.}

\item{clean_names}{If \code{TRUE} (default), converts column names to \code{snake_case}.}
}
\value{
A \link[tibble:tibble-package]{tibble}.
}
\description{
Retrieve school transport services overseen by the Court of Accounts.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/introspection.R
\name{tce_endpoint}
\alias{tce_endpoint}
\title{Get the catalog entry for a given endpoint}
\usage{
tce_endpoint(endpoint)
}
\arguments{
\item{endpoint}{Endpoint name (e.g., \code{"Municipios"}, \code{"LicitacaoUG"},
\code{"Contratos"}). Use \code{\link[=tce_catalog]{tce_catalog()}} to see all available endpoints.}
}
\value{
A list with endpoint metadata, including \code{entrada} (input
parameters) and \code{saida} (output fields) when available.
}
\description{
Returns the full metadata for a single endpoint, including its input
parameters (\code{entrada}) and output fields (\code{saida}). Useful for
programmatic access to endpoint metadata.
}
\examples{
tce_endpoint("Contratos")
}
\seealso{
\code{\link[=tce_catalog]{tce_catalog()}}, \code{\link[=tce_params]{tce_params()}}, \code{\link[=tce_fields]{tce_fields()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print-introspection.R
\name{print.tce_params_tbl}
\alias{print.tce_params_tbl}
\title{Print method for \code{tce_params()} output}
\usage{
\method{print}{tce_params_tbl}(x, ...)
}
\arguments{
\item{x}{An object returned by \code{\link[=tce_params]{tce_params()}}.}

\item{...}{Passed to the next print method.}
}
\value{
\code{x}, invisibly.
}
\description{
Print method for \code{tce_params()} output
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/request.R
\name{tce_request}
\alias{tce_request}
\title{Perform a request to the TCE-PE Open Data API}
\usage{
tce_request(
  endpoint,
  ...,
  clean_names = TRUE,
  max_tries = 3,
  progress = getOption("tceper.progress", TRUE),
  verbose = getOption("tceper.verbose", FALSE)
)
}
\arguments{
\item{endpoint}{A character string with the API method name (e.g.,
\code{"Contratos"}, \code{"DespesasEstaduais"}). See
\url{https://sistemas.tcepe.tc.br/DadosAbertos/Exemplo!listar} for the
full list of available endpoints.}

\item{...}{Named query parameters. You can use either the API names
(\code{CamelCase}/\code{UPPERCASE}) or the snake_case aliases shown by
\code{\link[=tce_params]{tce_params()}}. \code{NULL} values are silently dropped.}

\item{clean_names}{If \code{TRUE} (default), column names are converted to
\code{snake_case} via \code{\link[janitor:clean_names]{janitor::clean_names()}}. Set to \code{FALSE} to keep the
original column names returned by the API.}

\item{max_tries}{Maximum number of retry attempts for failed requests
(default: \code{3}). Passed to \code{\link[httr2:req_retry]{httr2::req_retry()}}.}

\item{progress}{If \code{TRUE}, displays progress messages via \link[cli:cli]{cli}. Defaults
to the value of \code{getOption("tceper.progress")}, or \code{TRUE} if unset.}

\item{verbose}{If \code{TRUE}, prints the final API URL, parameters, and
response metadata (status, content-type, size, elapsed time). Defaults
to \code{getOption("tceper.verbose", FALSE)}.}
}
\value{
A \link[tibble:tibble-package]{tibble} with the response content.
Returns a zero-row tibble if no records are found.
}
\description{
Low-level function that sends a GET request to the TCE-PE Open Data API
and returns the response content as a \link[tibble:tibble-package]{tibble}.
Higher-level wrapper functions such as \code{\link[=tce_contracts]{tce_contracts()}},
\code{\link[=tce_state_expenditures]{tce_state_expenditures()}}, etc. call this function internally.

You can use \code{tce_request()} directly to access any API endpoint,
including those that do not yet have a dedicated wrapper.
}
\details{
\subsection{API base URL}{

\verb{https://sistemas.tcepe.tc.br/DadosAbertos/}
}

\subsection{Network access}{

The API host only accepts connections from \strong{Brazilian IP addresses}.
Calls from outside Brazil are dropped at the network level and surface
as a connection or TLS error. The offline discovery functions
(\code{\link[=tce_catalog]{tce_catalog()}}, \code{\link[=tce_params]{tce_params()}}, \code{\link[=tce_fields]{tce_fields()}}) do not require
network access and work anywhere.
}

\subsection{URL construction}{

The TCE-PE API uses Struts2-style URLs where \code{!} invokes a method
(e.g. \verb{Contratos!json}). The URL is built with \code{paste0()} to preserve
the literal \code{!} -- using \code{httr2::req_url_path_append()} would encode it
as \verb{\%21}, causing the server to ignore all query parameters.
}

\subsection{Response limit}{

The API returns at most \strong{100,000 records} per request. When this limit
is reached, a warning is issued suggesting the use of additional filters.
}

\subsection{Encoding}{

The TCE-PE backend runs on \strong{ISO-8859-1} (Latin-1), not UTF-8. The
package handles this transparently in both directions:
\itemize{
\item \strong{Response body} is decoded as Latin-1 and returned as UTF-8 R
strings, so accented Portuguese characters come through cleanly
(\code{São}, \code{Município}, \code{Conceição}, etc.).
\item \strong{Query parameters} are transcoded UTF-8 -> Latin-1 before
percent-encoding, so filtering by accented values works as
expected (e.g. \code{tce_municipalities(municipio = "São José da Coroa Grande")}).
You can pass any R string with accents -- the conversion is
automatic.
}
}

\subsection{Options}{
\itemize{
\item \code{tceper.progress} -- set to \code{FALSE} to suppress all progress messages.
\item \code{tceper.verbose} -- set to \code{TRUE} to always print request/response details.
\item \code{tceper.cache_ttl} -- cache time-to-live in seconds (default: \code{3600}).
}
}
}
\examples{
\dontrun{
# Not run: these examples reach the live TCE-PE API, which only accepts
# connections from Brazilian IP addresses and is therefore unreachable
# from CRAN check machines and most CI runners.

# Direct query to any endpoint
tce_request("Contratos", CodigoEfiscoUG = "510101")

# Show full request/response details
tce_request("Contratos", CodigoEfiscoUG = "510101", verbose = TRUE)

# Keep original column names
tce_request("Municipios", clean_names = FALSE)
}
}
